# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
"Screen"
import calendar
import collections
import datetime
import functools
import gettext
import json
import logging
import urllib.parse
import xml.dom.minidom
from operator import itemgetter

from gi.repository import GLib, Gtk

from tryton.action import Action
from tryton.common import (
    MODELACCESS, RPCContextReload, RPCException, RPCExecute, node_attributes,
    sur, warning)
from tryton.common.domain_inversion import canonicalize
from tryton.common.domain_parser import DomainParser
from tryton.config import CONFIG
from tryton.gui.window.infobar import InfoBar
from tryton.gui.window.view_form.model.group import Group
from tryton.gui.window.view_form.view import View
from tryton.gui.window.view_form.view.screen_container import ScreenContainer
from tryton.jsonrpc import JSONEncoder
from tryton.pyson import PYSONDecoder
from tryton.rpc import clear_cache

_ = gettext.gettext
logger = logging.getLogger(__name__)


class Screen:
    "Screen"

    # Width of tree columns per model
    # It is shared with all connection but it is the price for speed.
    tree_column_width = collections.defaultdict(lambda: {})
    tree_column_optional = {}

    def __init__(self, model_name, **attributes):
        context = attributes.get('context', {})
        self._current_domain = []
        self.limit = attributes.get('limit', CONFIG['client.limit'])
        self.offset = 0
        self.windows = []

        self.readonly = attributes.get('readonly', False)
        if not (MODELACCESS[model_name]['write']
                or MODELACCESS[model_name]['create']):
            self.readonly = True
        self.search_count = 0
        if not attributes.get('row_activate'):
            self.row_activate = self.default_row_activate
        else:
            self.row_activate = attributes['row_activate']
        self.domain = attributes.get('domain', [])
        self.context_domain = attributes.get('context_domain')
        self.size_limit = None
        self.views_preload = attributes.get('views_preload', {})
        self.model_name = model_name
        self.views = []
        self.view_ids = attributes.get('view_ids', [])[:]
        self.parent = None
        self.parent_name = None
        self.exclude_field = attributes.get('exclude_field')
        self.filter_widget = None
        self.tree_states = collections.defaultdict(
            lambda: collections.defaultdict(lambda: None))
        self.tree_states_done = set()
        self.__group = None
        self.__current_record = None
        self.new_group(context or {})
        self.current_record = None
        self.screen_container = ScreenContainer(
            self, attributes.get('tab_domain'))
        self.screen_container.alternate_view = attributes.get(
            'alternate_view', False)
        self.widget = self.screen_container.widget_get()
        self.breadcrumb = attributes.get('breadcrumb') or []

        self.context_screen = None
        if attributes.get('context_model'):
            self.context_screen = Screen(
                attributes['context_model'], mode=['form'], context=context)
            self.context_screen.new()
            context_widget = self.context_screen.widget

            def walk_descendants(widget):
                yield widget
                if not hasattr(widget, 'get_children'):
                    return
                for child in widget.get_children():
                    for widget in walk_descendants(child):
                        yield widget

            for widget in reversed(list(walk_descendants(context_widget))):
                if isinstance(widget, Gtk.Entry):
                    widget.connect_after(
                        'activate', self.screen_container.activate)
                elif isinstance(widget, Gtk.CheckButton):
                    widget.connect_after(
                        'toggled', self.screen_container.activate)

            def remove_bin(widget):
                assert isinstance(widget, (Gtk.ScrolledWindow, Gtk.Viewport))
                parent = widget.get_parent()
                parent.remove(widget)
                child = widget.get_child()
                while isinstance(child, (Gtk.ScrolledWindow, Gtk.Viewport)):
                    child = child.get_child()
                child.get_parent().remove(child)
                parent.add(child)
                return child

            # Remove first level Viewport and ScrolledWindow to fill the Vbox
            remove_bin(self.context_screen.screen_container.viewport)
            if self.context_screen.current_view:
                remove_bin(
                    self.context_screen.current_view.widget.get_children()[0])

            self.screen_container.filter_vbox.pack_start(
                context_widget, expand=False, fill=True, padding=0)
            self.screen_container.filter_vbox.reorder_child(
                context_widget, 0)
            self.context_screen.widget.show()

        self.__current_view = 0
        self.search_value = attributes.get('search_value')
        self.fields_view_tree = {}
        self.order = self.default_order = attributes.get('order')
        self.view_to_load = []
        self._domain_parser = {}
        self.pre_validate = False
        mode = attributes.get('mode')
        if mode is None:
            mode = ['tree', 'form']
        self.view_to_load = mode[:]
        if self.view_ids or self.view_to_load:
            self.switch_view()

    def __repr__(self):
        return '<Screen %s at %s>' % (self.model_name, id(self))

    @property
    def readonly(self):
        return (self.__readonly
            or any(r.readonly for r in self.selected_records))

    @readonly.setter
    def readonly(self, value):
        self.__readonly = value

    @property
    def deletable(self):
        return all(r.deletable for r in self.selected_records)

    @property
    def count_limit(self):
        return self.limit * 100 + self.offset

    def search_active(self, active=True):
        if active and not self.parent:
            self.screen_container.show_filter()
        else:
            self.screen_container.hide_filter()

    @property
    def domain_parser(self):
        view_id = self.current_view.view_id if self.current_view else None

        if view_id in self._domain_parser:
            return self._domain_parser[view_id]

        if view_id not in self.fields_view_tree:
            context = self.context
            context['view_tree_width'] = CONFIG['client.save_tree_width']
            try:
                self.fields_view_tree[view_id] = view_tree = RPCExecute(
                    'model', self.model_name, 'fields_view_get', False, 'tree',
                    context=context)
            except RPCException:
                view_tree = {
                    'fields': {},
                    }
        else:
            view_tree = self.fields_view_tree[view_id]

        fields = view_tree['fields'].copy()
        for name in fields:
            if fields[name]['type'] not in {
                    'selection', 'multiselection', 'reference'}:
                continue
            if isinstance(fields[name]['selection'], (tuple, list)):
                continue
            props = fields[name] = fields[name].copy()
            props['selection'] = self.get_selection(props)

        if 'arch' in view_tree:
            # Filter only fields in XML view
            xml_dom = xml.dom.minidom.parseString(view_tree['arch'])
            root_node, = xml_dom.childNodes
            ofields = collections.OrderedDict()
            for node in root_node.childNodes:
                if node.nodeName != 'field':
                    continue
                attributes = node_attributes(node)
                name = attributes['name']
                # If a field is defined multiple times in the XML,
                # take only the first definition
                if name in ofields:
                    continue
                ofields[name] = fields[name]
                for attr in ['string', 'factor']:
                    if attributes.get(attr):
                        ofields[name][attr] = attributes[attr]
                symbol = attributes.get('symbol')
                if symbol and symbol not in ofields:
                    ofields[symbol] = fields[symbol]
            fields = ofields

        # Add common fields
        for name, string, type_ in (
                ('id', _('ID'), 'integer'),
                ('create_uid', _('Created by'), 'many2one'),
                ('create_date', _('Created at'), 'datetime'),
                ('write_uid', _('Edited by'), 'many2one'),
                ('write_date', _('Edited at'), 'datetime'),
                ):
            if name not in fields:
                fields[name] = {
                    'string': string,
                    'name': name,
                    'type': type_,
                    }
                if type_ == 'datetime':
                    fields[name]['format'] = '"%H:%M:%S"'

        domain_parser = DomainParser(fields, self.context)
        self._domain_parser[view_id] = domain_parser
        return domain_parser

    def get_selection(self, props):
        try:
            change_with = props.get('selection_change_with')
            if change_with:
                selection = RPCExecute('model', self.model_name,
                    props['selection'], dict((p, None) for p in change_with))
            else:
                selection = RPCExecute('model', self.model_name,
                    props['selection'])
        except RPCException:
            selection = []
        selection.sort(key=itemgetter(1))
        return selection

    def search_prev(self, search_string):
        if self.limit:
            self.offset = max(self.offset - self.limit, 0)
        self.search_filter(search_string=search_string)

    def search_next(self, search_string):
        if self.limit:
            self.offset += self.limit
        self.search_filter(search_string=search_string)

    def search_complete(self, search_string):
        return list(self.domain_parser.completion(search_string))

    def search_filter(self, search_string=None, only_ids=False):
        if self.context_screen and not only_ids:
            context_record = self.context_screen.current_record
            if not context_record.validate():
                self.clear()
                self.context_screen.display(set_cursor=True)
                return False
            context = self.local_context
            screen_context = self.context_screen.get_on_change_value()
            screen_context.pop('id')
            context.update(screen_context)
            self.new_group(context)

        domain = self.search_domain(search_string, True)
        if (canonicalized := canonicalize(domain)) != self._current_domain:
            self._current_domain = canonicalized
            self.offset = 0

        context = self.context
        if (self.screen_container.but_active.props.visible
                and self.screen_container.but_active.get_active()):
            context['active_test'] = False
        ids = []
        while True:
            try:
                ids = RPCExecute('model', self.model_name, 'search', domain,
                    self.offset, self.limit, self.order, context=context)
            except RPCException:
                break
            if ids or self.offset <= 0:
                break
            self.offset = max(self.offset - self.limit, 0)
        if not only_ids:
            if self.limit is not None and len(ids) == self.limit:
                try:
                    self.search_count = RPCExecute(
                        'model', self.model_name, 'search_count',
                        domain, 0, self.count_limit, context=context)
                except RPCException:
                    self.search_count = 0
            else:
                self.search_count = len(ids)
        self.screen_container.but_prev.set_sensitive(bool(self.offset))
        if (self.limit is not None
                and len(ids) == self.limit
                and self.search_count > self.limit + self.offset):
            self.screen_container.but_next.set_sensitive(True)
        else:
            self.screen_container.but_next.set_sensitive(False)
        if only_ids:
            return ids
        self.clear()
        self.load(ids)
        self.count_tab_domain()
        return bool(ids)

    def search_domain(self, search_string=None, set_text=False, with_tab=True):
        domain = []
        # Test first parent to avoid calling unnecessary domain_parser
        if not self.parent and self.domain_parser:
            if search_string is not None:
                domain = self.domain_parser.parse(search_string)
            else:
                domain = self.search_value
                self.search_value = None
            if set_text:
                self.screen_container.set_text(
                    self.domain_parser.string(domain))
        else:
            domain = [('id', 'in', [x.id for x in self.group])]

        if domain:
            if self.domain:
                domain = ['AND', domain, self.domain]
        else:
            domain = self.domain

        if (self.screen_container.but_active.props.visible
                and self.screen_container.but_active.get_active()):
            if domain:
                domain = [domain, ('active', '=', False)]
            else:
                domain = [('active', '=', False)]
        if self.current_view and self.current_view.view_type == 'calendar':
            if domain:
                domain = ['AND', domain, self.current_view.current_domain()]
            else:
                domain = self.current_view.current_domain()
        if self.context_domain:
            decoder = PYSONDecoder(self.context)
            domain = ['AND', domain, decoder.decode(self.context_domain)]
        if with_tab:
            tab_domain = self.screen_container.get_tab_domain()
            if tab_domain:
                domain = ['AND', domain, tab_domain]
        return domain

    def count_tab_domain(self, current=False):
        def set_tab_counter(count, idx):
            try:
                count = count()
            except RPCException:
                count = None
            self.screen_container.set_tab_counter(count, idx)
        screen_domain = self.search_domain(
            self.screen_container.get_text(), with_tab=False)
        index = self.screen_container.get_tab_index()
        for idx, (name, domain, count) in enumerate(
                self.screen_container.tab_domain):
            if not count or (current and idx != index):
                continue
            domain = ['AND', domain, screen_domain]
            set_tab_counter(lambda: None, idx)
            RPCExecute('model', self.model_name,
                'search_count', domain, 0, 1000, context=self.context,
                callback=functools.partial(set_tab_counter, idx=idx))

    @property
    def context(self):
        context = self.group.context
        if self.context_screen:
            context['context_model'] = self.context_screen.model_name
        return context

    @property
    def local_context(self):
        context = self.group.local_context
        if self.context_screen:
            context['context_model'] = self.context_screen.model_name
        return context

    def __get_group(self):
        return self.__group

    def __set_group(self, group):
        fields = {}
        fields_views = {}
        if self.group is not None:
            for name, field in self.group.fields.items():
                fields[name] = field.attrs
                fields_views[name] = field.views
            if self in self.group.screens:
                self.group.screens.remove(self)
            group.on_write.update(self.group.on_write)
        self.tree_states_done.clear()
        self.__group = group
        self.group.screens.append(self)
        self.parent = group.parent
        self.parent_name = group.parent_name
        if self.parent:
            self.filter_widget = None
            self.order = None
        self.__group.add_fields(fields)
        self.current_record = None
        for name, views in fields_views.items():
            self.__group.fields[name].views.update(views)
        self.__group.exclude_field = self.exclude_field

    group = property(__get_group, __set_group)

    def new_group(self, context=None):
        context = context if context is not None else self.context
        self.group = Group(self.model_name, {}, domain=self.domain,
            context=context, readonly=self.__readonly)

    def group_list_changed(self, group, action, *args):
        for view in self.views:
            if hasattr(view, 'group_list_changed'):
                view.group_list_changed(group, action, *args)

    def record_modified(self, display=True):
        for window in self.windows:
            if hasattr(window, 'record_modified'):
                window.record_modified()
        if display:
            self.display()

    def record_notify(self, notifications):
        for window in self.windows:
            if isinstance(window, InfoBar):
                window.info_bar_refresh('notification')
                for type_, message in notifications:
                    type_ = {
                        'info': Gtk.MessageType.INFO,
                        'warning': Gtk.MessageType.WARNING,
                        'error': Gtk.MessageType.ERROR,
                        }.get(type_, Gtk.MessageType.WARNING)
                    window.info_bar_add(message, type_, 'notification')

    def record_message(self, position, size, max_size, record_id):
        for window in self.windows:
            if hasattr(window, 'record_message'):
                window.record_message(position, size, max_size, record_id)

    def record_saved(self):
        for window in self.windows:
            if hasattr(window, 'record_saved'):
                window.record_saved()

    def update_resources(self, resources):
        for window in self.windows:
            if hasattr(window, 'update_resources'):
                window.update_resources(resources)

    def has_update_resources(self):
        return any(hasattr(w, 'update_resources') for w in self.windows)

    def __get_current_record(self):
        if (self.__current_record is not None
                and self.__current_record.group is None):
            self.__current_record = None
        return self.__current_record

    def __set_current_record(self, record):
        self.__current_record = record
        if record:
            try:
                pos = self.group.index(record) + self.offset + 1
            except ValueError:
                # XXX offset?
                pos = record.get_index_path()
        else:
            pos = 0
        self.record_message(
            pos, len(self.group) + self.offset,
            self.search_count, record and record.id)
        self.update_resources(record.resources if record else None)
        # update resources after 1 second
        GLib.timeout_add(1000, self._update_resources, record)

    current_record = property(__get_current_record, __set_current_record)

    def _update_resources(self, record):
        if (record
                and record == self.current_record
                and self.has_update_resources()):
            self.update_resources(record.get_resources())
        return False

    def destroy(self):
        self.windows.clear()
        for view in self.views:
            view.destroy()
        del self.views[:]
        self.group.destroy()

    def default_row_activate(self):
        if (self.current_view
                and self.current_view.view_type == 'tree'
                and int(self.current_view.attributes.get('keyword_open', 0))):
            return Action.exec_keyword('tree_open', {
                'model': self.model_name,
                'id': self.current_record.id if self.current_record else None,
                'ids': [r.id for r in self.selected_records],
                }, context=self.local_context, warning=False)
        else:
            if not self.modified():
                self.switch_view(view_type='form')
            return True

    @property
    def number_of_views(self):
        return len(self.views) + len(self.view_to_load)

    @property
    def view_index(self):
        return self.__current_view

    @property
    def next_view_type(self):
        views = self.views + self.view_to_load
        next_view_index = (self.view_index + 1) % len(views)
        next_view = views[next_view_index]
        if not isinstance(next_view, str):
            next_view = next_view.view_type
        return next_view

    def switch_view(
            self, view_type=None, view_id=None, creatable=None, display=True):
        if view_id is not None:
            view_id = int(view_id)
        if self.current_view:
            self.current_view.set_value()
            if (self.current_record
                    and self.current_record not in self.current_record.group):
                self.current_record = None
            fields = self.current_view.get_fields()
            if (self.current_record and self.current_view.editable
                    and not self.current_record.validate(fields)):
                self.screen_container.set(self.current_view.widget)
                self.set_cursor()
                self.current_view.display()
                return

        def found():
            if not self.current_view:
                return False
            result = True
            if view_type is not None:
                result &= self.current_view.view_type == view_type
            if view_id is not None:
                result &= self.current_view.view_id == view_id
            if creatable is not None:
                result &= self.current_view.creatable == creatable
            return result
        for i in range(len(self.views) + len(self.view_to_load)):
            if len(self.view_to_load):
                self.load_view_to_load()
                self.__current_view = len(self.views) - 1
            elif (view_id is not None
                    and view_id not in {v.view_id for v in self.views}):
                self.add_view_id(view_id, view_type)
                self.__current_view = len(self.views) - 1
                break
            else:
                self.__current_view = ((self.__current_view + 1)
                        % len(self.views))
            if found():
                break
        self.screen_container.set(self.current_view.widget)
        if display:
            self.display()
            # Postpone set of the cursor to ensure widgets are allocated
            GLib.idle_add(self.set_cursor)

    def load_view_to_load(self):
        if len(self.view_to_load):
            if self.view_ids:
                view_id = self.view_ids.pop(0)
            else:
                view_id = None
            view_type = self.view_to_load.pop(0)
            self.add_view_id(view_id, view_type)

    def add_view_id(self, view_id, view_type):
        if view_id and str(view_id) in self.views_preload:
            view = self.views_preload[str(view_id)]
        elif not view_id and view_type in self.views_preload:
            view = self.views_preload[view_type]
        else:
            context = self.context
            context['view_tree_width'] = CONFIG['client.save_tree_width']
            try:
                view = RPCExecute(
                    'model', self.model_name, 'fields_view_get', view_id,
                    view_type, context=context)
            except RPCException:
                return
        return self.add_view(view)

    def add_view(self, view):
        arch = view['arch']
        fields = view['fields']
        view_id = view['view_id']

        xml_dom = xml.dom.minidom.parseString(arch)
        root, = xml_dom.childNodes
        if root.tagName == 'tree':
            self.fields_view_tree[view_id] = view

        # Ensure that loading is always lazy for fields on form view
        # and always eager for fields on tree or graph view
        if root.tagName == 'form':
            loading = 'lazy'
        else:
            loading = 'eager'
        for field in fields:
            if field not in self.group.fields or loading == 'eager':
                fields[field]['loading'] = loading
            else:
                fields[field]['loading'] = \
                    self.group.fields[field].attrs['loading']
        self.group.add_fields(fields)
        for field in fields:
            self.group.fields[field].views.add(view_id)
        view = View.parse(
            self, view_id, view['type'], xml_dom, view.get('field_childs'))
        self.views.append(view)

        return view

    def new(self, default=True, defaults=None):
        previous_view = self.current_view
        if self.current_view and self.current_view.view_type == 'calendar':
            selected_date = self.current_view.get_selected_date()
        if self.current_view and not self.current_view.creatable:
            self.switch_view(creatable=True)
            if not self.current_view.creatable:
                return None
        if self.current_record:
            group = self.current_record.group
        else:
            group = self.group
        record = group.new(default, defaults=defaults)
        group.add(record, self.new_position)
        if previous_view.view_type == 'calendar':
            previous_view.set_default_date(record, selected_date)
        self.current_record = record
        self.display()
        # Postpone set of the cursor to ensure widgets are allocated
        GLib.idle_add(self.set_cursor, True)
        return self.current_record

    @property
    def new_position(self):
        if self.order is not None:
            order = self.order
        else:
            order = self.default_order
        if order:
            for oexpr, otype in order:
                if oexpr == 'id' and otype:
                    if otype.startswith('DESC'):
                        return 0
                    elif otype.startswith('ASC'):
                        return -1
        if self.parent:
            return -1
        else:
            return 0

    def set_on_write(self, func_name):
        if func_name:
            self.group.on_write.add(func_name)

    def cancel_current(self, initial_value=None):
        if self.current_record:
            self.current_record.cancel()
            if self.current_record.id < 0:
                if initial_value is not None:
                    self.current_record.reset(initial_value)
                else:
                    self.remove(records=[self.current_record])

    def save_current(self):
        if not self.current_record:
            if (self.current_view
                    and self.current_view.view_type == 'tree'
                    and len(self.group)):
                self.current_record = self.group[0]
            else:
                return True
        saved = False
        record_id = None
        if self.current_view:
            self.current_view.set_value()
            fields = self.current_view.get_fields()
        path = self.current_record.get_path(self.group)
        if self.current_view and self.current_view.view_type == 'tree':
            # False value must be not saved
            saved = all((
                    x is not False and x >= 0
                    for x in self.group.save()))
            record_id = self.current_record.id if self.current_record else None
        elif self.current_record.validate(fields):
            record_id = self.current_record.save(force_reload=True)
            # False value must be not saved
            saved = record_id is not False and record_id >= 0
        elif self.current_view:
            self.set_cursor()
            self.current_view.display()
            return False
        if path and record_id:
            path = path[:-1] + ((path[-1][0], record_id),)
        self.current_record = self.group.get_by_path(path)
        self.display()
        self.record_saved()
        return saved

    def __get_current_view(self):
        if not len(self.views):
            return None
        return self.views[self.__current_view]

    current_view = property(__get_current_view)

    def set_cursor(self, new=False, reset_view=True):
        current_view = self.current_view
        if not current_view:
            return
        elif current_view.view_type in ('tree', 'form', 'list-form'):
            current_view.set_cursor(new=new, reset_view=reset_view)

    def get(self):
        if not self.current_record:
            return None
        if self.current_view:
            self.current_view.set_value()
        return self.current_record.get()

    def get_on_change_value(self):
        if not self.current_record:
            return None
        if self.current_view:
            self.current_view.set_value()
        return self.current_record.get_on_change_value()

    def modified(self):
        if self.current_view and self.current_view.view_type != 'tree':
            if self.current_record:
                if self.current_record.modified or self.current_record.id < 0:
                    return True
        else:
            for record in self.group:
                if record.modified or record.id < 0:
                    return True
        if self.current_view and self.current_view.modified:
            return True
        return False

    def reload(self, ids, written=False):
        self.group.reload(ids)
        if written:
            self.group.written(ids)
        if self.parent:
            self.parent.root_parent.reload()
        self.display()

    def unremove(self):
        records = self.selected_records
        for record in records:
            self.group.unremove(record)

    def remove(self, delete=False, remove=False, force_remove=False,
            records=None):
        records = records or self.selected_records
        if not records:
            return
        if delete:
            # Must delete children records before parent
            records.sort(key=lambda r: r.depth, reverse=True)
            if not self.group.delete(records):
                return False

        for record in records:
            # set current model to None to prevent __select_changed
            # to save the previous_model as it can be already deleted.
            self.current_record = None
            record.group.remove(
                record, remove=remove, modified=False,
                force_remove=force_remove)
        # set current_record to None to prevent __select_changed
        # to set deleted record as current_record
        self.current_record = None
        # call only once
        record.set_modified()

        if delete:
            for record in records:
                if record in record.group.record_deleted:
                    record.group.record_deleted.remove(record)
                if record in record.group.record_removed:
                    record.group.record_removed.remove(record)
                if record.parent:
                    # Save parent without deleted children
                    record.parent.save(force_reload=False)
                record.destroy()

        self.current_record = None
        self.set_cursor()
        self.display()
        return True

    def copy(self):
        ids = [r.id for r in self.selected_records]
        try:
            new_ids = RPCExecute('model', self.model_name, 'copy', ids, {},
                context=self.context)
        except RPCException:
            return False
        self.group.load(new_ids, position=self.new_position)
        if new_ids:
            self.current_record = self.group.get(new_ids[0])
        self.display(set_cursor=True)
        return True

    def set_tree_state(self):
        view = self.current_view
        if not view:
            return
        if view.view_type not in {'tree', 'form', 'list-form'}:
            return
        if id(view) in self.tree_states_done:
            return
        if view.view_type == 'form' and self.tree_states_done:
            return
        if (view.view_type in {'tree', 'list-form'}
                and not int(view.attributes.get('tree_state', False))):
            # Mark as done to not set later when the view_type change
            self.tree_states_done.add(id(view))
        parent = self.parent.id if self.parent else None
        if parent is not None and parent < 0:
            return
        state = self.tree_states[parent][view.children_field]
        if state:
            expanded_nodes, selected_nodes = state
        else:
            expanded_nodes, selected_nodes = [], []
        if view.view_type in {'tree', 'list-form'}:
            if (state is None
                    and CONFIG['client.save_tree_state']
                    and int(view.attributes.get('tree_state', False))):
                json_domain = self.get_tree_domain(parent)
                try:
                    expanded_nodes, selected_nodes = RPCExecute('model',
                        'ir.ui.view_tree_state', 'get',
                        self.model_name, json_domain,
                        view.children_field)
                    expanded_nodes = json.loads(expanded_nodes)
                    selected_nodes = json.loads(selected_nodes)
                except RPCException:
                    logger.warn(
                        'Unable to get view tree state for %s',
                        self.model_name)
                self.tree_states[parent][view.children_field] = (
                    expanded_nodes, selected_nodes)
            if view.view_type == 'tree':
                view.expand_nodes(expanded_nodes)
            view.select_nodes(selected_nodes)
        else:
            if selected_nodes and not self.current_record:
                record = None
                for node in selected_nodes[0]:
                    new_record = self.group.get(node)
                    if node < 0 and -node < len(self.group):
                        # Negative id is the index of the new record
                        new_record = self.group[-node]
                    if not new_record:
                        break
                    else:
                        record = new_record
                if record and record != self.current_record:
                    self.current_record = record
                    # Force a display of the view to synchronize the
                    # widgets with the new record
                    view.display()
        self.tree_states_done.add(id(view))

    def save_tree_state(self, store=True):
        parent = self.parent.id if self.parent else None
        for view in self.views:
            if view.view_type == 'form':
                for widgets in view.widgets.values():
                    for widget in widgets:
                        if hasattr(widget, 'screen'):
                            widget.screen.save_tree_state(store)
            if view == self.current_view and view.view_type == 'form':
                if self.current_record:
                    path = self.current_record.id
                    if path < 0:
                        path = -self.current_record.group.index(
                            self.current_record)
                    self.tree_states[parent][view.children_field] = (
                        [], [[path]])
            elif view.view_type in {'tree', 'list-form'}:
                if view.view_type == 'tree':
                    view.save_width()
                    paths = view.get_expanded_paths()
                else:
                    paths = []
                selected_paths = view.get_selected_paths()
                if view == self.current_view:
                    self.tree_states[parent][view.children_field] = (
                        paths, selected_paths)
                if (store
                        and int(view.attributes.get('tree_state', False))
                        and CONFIG['client.save_tree_state']):
                    json_domain = self.get_tree_domain(parent)
                    json_paths = json.dumps(paths, separators=(',', ':'))
                    json_selected_path = json.dumps(
                        selected_paths, separators=(',', ':'))
                    try:
                        RPCExecute('model', 'ir.ui.view_tree_state', 'set',
                            self.model_name, json_domain, view.children_field,
                            json_paths, json_selected_path,
                            process_exception=False)
                        clear_cache('model.ir.ui.view_tree_state.get')
                    except Exception:
                        logger.warn(
                            _('Unable to set view tree state'), exc_info=True)

    def get_tree_domain(self, parent):
        if parent:
            domain = (self.domain + [(self.exclude_field, '=', parent)])
        else:
            domain = self.domain
        json_domain = json.dumps(
            domain, cls=JSONEncoder, separators=(',', ':'))
        return json_domain

    def load(self, ids, set_cursor=True, modified=False, position=-1):
        self.group.load(ids, modified=modified, position=position)
        if self.current_view:
            self.current_view.reset()
        self.current_record = None
        self.display(set_cursor=set_cursor)

    def display(self, set_cursor=False):
        if self.views and self.current_view:
            self.search_active(self.current_view.view_type
                in ('tree', 'graph', 'calendar'))
            for view in self.views:
                # Always display tree view to update model
                # because view can be used even if it is not shown
                # like for save_tree_state
                if (view == self.current_view
                        or view.view_type == 'tree'
                        or view.widget.get_parent()):
                    view.display()
            self.current_view.widget.set_sensitive(
                bool(self.group
                    or (self.current_view.view_type != 'form')
                    or self.current_record))
            if self.current_view.view_type == 'tree':
                view_tree = self.fields_view_tree.get(
                    self.current_view.view_id, {})
                if 'active' in view_tree['fields']:
                    self.screen_container.but_active.show()
                else:
                    self.screen_container.but_active.hide()
            else:
                self.screen_container.but_active.hide()
            if set_cursor:
                self.set_cursor(reset_view=False)
        self.set_tree_state()
        # Force record_message
        self.current_record = self.current_record

    def _get_next_record(self, test=False):
        view = self.current_view
        if (view
                and view.view_type in {'tree', 'form'}
                and self.current_record
                and self.current_record.group):
            group = self.current_record.group
            record = self.current_record
            while group:
                children = record.children_group(view.children_field)
                if children:
                    record = children[0]
                    break
                idx = group.index(record) + 1
                if idx < len(group):
                    record = group[idx]
                    break
                parent = record.parent
                if not parent or record.model_name != parent.model_name:
                    break
                next = parent.next.get(id(parent.group))
                while not next:
                    parent = parent.parent
                    if not parent:
                        break
                    next = parent.next.get(id(parent.group))
                if not next:
                    break
                record = next
                break
            return record
        elif (view
                and view.view_type == 'list-form'
                and len(self.group)
                and self.current_record in self.group):
            idx = self.group.index(self.current_record)
            if 0 <= idx < len(self.group) - 1:
                return self.group[idx + 1]
        elif view and view.view_type == 'calendar':
            record = self.current_record
            goocalendar = view.widgets.get('goocalendar')
            if goocalendar:
                date = goocalendar.selected_date
                year = date.year
                month = date.month
                start = datetime.datetime(year, month, 1)
                nb_days = calendar.monthrange(year, month)[1]
                delta = datetime.timedelta(days=nb_days)
                end = start + delta
                events = goocalendar.event_store.get_events(start, end)
                events.sort()
                if not record:
                    if events:
                        return events[0].record
                    else:
                        return
                else:
                    for idx, event in enumerate(events):
                        if event.record == record:
                            next_id = idx + 1
                            if next_id < len(events):
                                return events[next_id].record
                            break
        else:
            return self.group[0] if len(self.group) else None

    def has_next(self):
        next_record = self._get_next_record(test=True)
        return next_record and next_record != self.current_record

    def display_next(self):
        view = self.current_view
        if view:
            view.set_value()
        self.set_cursor(reset_view=False)
        self.current_record = self._get_next_record()
        self.set_cursor(reset_view=False)
        if view:
            view.display()

    def _get_prev_record(self, test=False):
        view = self.current_view
        if (view
                and view.view_type in {'tree', 'form'}
                and self.current_record
                and self.current_record.group):
            group = self.current_record.group
            record = self.current_record
            idx = group.index(record) - 1
            if idx >= 0:
                record = group[idx]
                children = True
                while children:
                    children = record.children_group(view.children_field)
                    if children:
                        record = children[-1]
            else:
                parent = record.parent
                if parent and record.model_name == parent.model_name:
                    record = parent
            return record
        elif view and view.view_type == 'calendar':
            record = self.current_record
            goocalendar = view.widgets.get('goocalendar')
            if goocalendar:
                date = goocalendar.selected_date
                year = date.year
                month = date.month
                start = datetime.datetime(year, month, 1)
                nb_days = calendar.monthrange(year, month)[1]
                delta = datetime.timedelta(days=nb_days)
                end = start + delta
                events = goocalendar.event_store.get_events(start, end)
                events.sort()
                if not record:
                    if events:
                        return events[0].record
                    else:
                        return
                else:
                    for idx, event in enumerate(events):
                        if event.record == record:
                            prev_id = idx - 1
                            if prev_id >= 0:
                                return events[prev_id].record
                            break
        elif (view
                and view.view_type == 'list-form'
                and len(self.group)
                and self.current_record in self.group):
            idx = self.group.index(self.current_record)
            if 0 < idx <= len(self.group) - 1:
                return self.group[idx - 1]
        else:
            return self.group[-1] if len(self.group) else None

    def has_prev(self):
        prev_record = self._get_prev_record(test=True)
        return prev_record and prev_record != self.current_record

    def display_prev(self):
        view = self.current_view
        if view:
            view.set_value()
        self.set_cursor(reset_view=False)
        self.current_record = self._get_prev_record()
        self.set_cursor(reset_view=False)
        if view:
            view.display()

    def invalid_message(self, record=None):
        if record is None:
            record = self.current_record
        domain_string = _('"%s" is not valid according to its domain.')
        domain_parser = DomainParser(
            {n: f.attrs for n, f in record.group.fields.items()})
        fields = []
        for field, invalid in sorted(record.invalid_fields.items()):
            string = record.group.fields[field].attrs['string']
            if invalid == 'required' or invalid == [[field, '!=', None]]:
                fields.append(_('"%s" is required.') % string)
            elif invalid == 'domain':
                fields.append(domain_string % string)
            elif invalid == 'children':
                fields.append(_('The values of "%s" are not valid.') % string)
            else:
                if domain_parser.stringable(invalid):
                    fields.append(domain_parser.string(invalid))
                else:
                    fields.append(domain_string % string)
        if len(fields) > 5:
            fields = fields[:5] + ['...']
        return '\n'.join(fields)

    @property
    def selected_records(self):
        return self.current_view.selected_records if self.current_view else []

    @property
    def selected_paths(self):
        if self.current_view and self.current_view.view_type == 'tree':
            return self.current_view.get_selected_paths()

    @property
    def listed_records(self):
        if (self.current_view
                and self.current_view.view_type in {
                    'tree', 'calendar', 'list-form'}):
            return self.current_view.listed_records
        elif self.current_record:
            return [self.current_record]
        else:
            return []

    @property
    def listed_paths(self):
        if self.current_view and self.current_view.view_type == 'tree':
            return self.current_view.get_listed_paths()

    def clear(self):
        self.current_record = None
        self.group.clear()
        self.tree_states_done.clear()
        for view in self.views:
            view.reset()

    def on_change(self, fieldname, attr):
        self.current_record.on_change(fieldname, attr)
        self.display()

    def get_buttons(self):
        'Return active buttons for the current view'
        def is_active(record, button):
            if button.attrs.get('type', 'class') == 'instance':
                return False
            states = record.expr_eval(button.attrs.get('states', {}))
            return not (states.get('invisible') or states.get('readonly'))

        if not self.selected_records:
            return []

        buttons = self.current_view.get_buttons() if self.current_view else []

        for record in self.selected_records:
            buttons = [b for b in buttons if is_active(record, b)]
            if not buttons:
                break
        return buttons

    def button(self, button):
        'Execute button on the selected records'
        if self.current_view:
            self.current_view.set_value()
            fields = self.current_view.get_fields()
        for record in self.selected_records:
            domain = record.expr_eval(
                button.get('states', {})).get('pre_validate', [])
            if not record.validate(fields, pre_validate=domain):
                warning(self.invalid_message(record), _('Pre-validation'))
                self.display(set_cursor=True)
                if domain:
                    # Reset valid state with normal domain
                    record.validate(fields)
                return
        if button.get('confirm', False) and not sur(button['confirm']):
            return
        if button.get('type', 'class') == 'class':
            record_id = self.current_record.save(force_reload=False)
            if record_id is False or record_id < 0:
                return
        if button.get('type', 'class') == 'class':
            self._button_class(button)
        else:
            self._button_instance(button)

    def _button_instance(self, button):
        record = self.current_record
        args = record.expr_eval(button.get('change', []))
        values = record._get_on_change_args(args)
        try:
            changes = RPCExecute('model', self.model_name, button['name'],
                values, context=self.context)
        except RPCException:
            return
        record.set_on_change(changes)
        record.set_modified()

    def _button_class(self, button):
        ids = [r.id for r in self.selected_records]
        context = self.context
        context['_timestamp'] = {}
        for record in self.selected_records:
            context['_timestamp'].update(record.get_timestamp())
        try:
            action = RPCExecute('model', self.model_name, button['name'],
                ids, context=context)
        except RPCException:
            action = None
        self.reload(ids, written=True)
        self.record_saved()
        if isinstance(action, str):
            self.client_action(action)
        elif action:
            Action.execute(action, {
                    'model': self.model_name,
                    'id': self.current_record.id,
                    'ids': ids,
                    }, context=self.context, keyword=True)

    def client_action(self, action):
        access = MODELACCESS[self.model_name]
        if action == 'new':
            if access['create']:
                self.new()
        elif action == 'delete':
            if (access['delete']
                    and (self.current_record.deletable
                        if self.current_record else True)):
                self.remove(delete=not self.parent,
                    force_remove=not self.parent)
        elif action == 'remove':
            if access['write'] and access['read'] and self.parent:
                self.remove(remove=True)
        elif action == 'copy':
            if access['create']:
                self.copy()
        elif action == 'next':
            self.display_next()
        elif action == 'previous':
            self.display_prev()
        elif action == 'close':
            from tryton.gui import Main
            Main().sig_win_close()
        elif action.startswith('switch'):
            self.switch_view(*action.split(None, 2)[1:])
        elif action == 'reload':
            if (self.current_view
                    and self.current_view.view_type in [
                        'tree', 'graph', 'calendar']
                    and not self.parent):
                self.search_filter()
        elif action == 'reload menu':
            from tryton.gui import Main
            RPCContextReload(Main().sig_win_menu)
        elif action == 'reload context':
            RPCContextReload()

    def get_url(self, name=''):
        query_string = []
        if self.domain:
            query_string.append(('domain', json.dumps(
                        self.domain, cls=JSONEncoder, separators=(',', ':'))))
        context = self.local_context  # Avoid rpc context
        if context:
            query_string.append(('context', json.dumps(
                        context, cls=JSONEncoder, separators=(',', ':'))))
        if self.context_screen:
            query_string.append(
                ('context_model', self.context_screen.model_name))
        if name:
            query_string.append(
                ('name', json.dumps(name, separators=(',', ':'))))
        if self.screen_container.tab_domain:
            query_string.append(('tab_domain', json.dumps(
                        self.screen_container.tab_domain,
                        cls=JSONEncoder, separators=(',', ':'))))
        path = [CONFIG['login.db'], 'model', self.model_name]
        view_ids = [v.view_id for v in self.views] + self.view_ids
        if self.current_view and self.current_view.view_type != 'form':
            if self.search_value:
                search_value = self.search_value
            else:
                search_string = self.screen_container.get_text()
                search_value = self.domain_parser.parse(search_string)
            if search_value:
                query_string.append(('search_value', json.dumps(
                            search_value, cls=JSONEncoder,
                            separators=(',', ':'))))
        elif self.current_record and self.current_record.id > -1:
            path.append(str(self.current_record.id))
            if self.current_view:
                i = view_ids.index(self.current_view.view_id)
                view_ids = view_ids[i:] + view_ids[:i]
        if view_ids:
            query_string.append(('views', json.dumps(
                        view_ids, separators=(',', ':'))))
        query_string = urllib.parse.urlencode(query_string)
        return urllib.parse.urlunparse(('tryton',
                CONFIG['login.host'],
                '/'.join(path), query_string, '', ''))



# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
from gi.repository import Gtk

from tryton.common.entry_position import reset_position

from .widget import Widget


class TimeDelta(Widget):

    def __init__(self, view, attrs):
        super(TimeDelta, self).__init__(view, attrs)

        self.widget = Gtk.HBox()
        self.entry = self.mnemonic_widget = Gtk.Entry()
        self.entry.set_alignment(1.0)
        self.entry.set_property('activates_default', True)

        self.entry.connect('activate', self.sig_activate)
        self.entry.connect('focus-out-event', lambda x, y: self._focus_out())
        self.entry.connect('key-press-event', self.send_modified)
        self.widget.pack_start(self.entry, expand=True, fill=True, padding=0)

    @property
    def modified(self):
        if self.record and self.field:
            value = self.entry.get_text()
            return self.field.get_client(self.record) != value
        return False

    def set_value(self):
        value = self.entry.get_text()
        return self.field.set_client(self.record, value)

    def get_value(self):
        return self.entry.get_text()

    def display(self):
        super(TimeDelta, self).display()
        if not self.field:
            value = ''
        else:
            value = self.field.get_client(self.record)
        self.entry.set_text(value)
        reset_position(self.entry)

    def _readonly_set(self, value):
        super(TimeDelta, self)._readonly_set(value)
        self.entry.set_editable(not value)



# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
from gi.repository import GLib, Gtk

from tryton.common.selection import (
    PopdownMixin, SelectionMixin, selection_shortcuts)

from .widget import Widget


class Selection(Widget, SelectionMixin, PopdownMixin):

    def __init__(self, view, attrs):
        super(Selection, self).__init__(view, attrs)

        self.widget = Gtk.HBox(spacing=3)
        self.entry = Gtk.ComboBox(has_entry=True)
        child = self.mnemonic_widget = self.entry.get_child()
        child.set_property('activates_default', True)
        child.set_max_length(int(attrs.get('size', 0)))
        child.set_width_chars(self.default_width_chars)

        selection_shortcuts(self.entry)
        child.connect('activate', lambda *a: self._focus_out())
        child.connect('focus-out-event', lambda *a: self._focus_out())
        self.entry.connect('changed', self.changed)
        self.entry.connect('move-active', self._move_active)
        self.entry.connect(
            'scroll-event',
            lambda c, e: c.stop_emission_by_name('scroll-event'))
        self.widget.pack_start(self.entry, expand=True, fill=True, padding=0)

        self.selection = attrs.get('selection', [])[:]
        self.attrs = attrs
        self.init_selection()
        self.set_popdown(self.selection, self.entry)

    def changed(self, combobox):
        def focus_out():
            if combobox.props.window:
                self._focus_out()
        # Must be deferred because it triggers a display of the form
        GLib.idle_add(focus_out)

    def _move_active(self, combobox, scroll_type):
        if not combobox.get_child().get_editable():
            combobox.stop_emission_by_name('move-active')

    def _readonly_set(self, value):
        super(Selection, self)._readonly_set(value)
        self.entry.get_child().set_editable(not value)
        self.entry.set_button_sensitivity(
            Gtk.SensitivityType.OFF if value else Gtk.SensitivityType.AUTO)

    def get_value(self):
        if not self.entry.get_child():  # entry is destroyed
            return
        return self.get_popdown_value(self.entry)

    @property
    def modified(self):
        if self.record and self.field:
            return self.field.get(self.record) != self.get_value()
        return False

    def set_value(self):
        value = self.get_value()
        if 'relation' in self.attrs and value:
            value = (value, self.get_popdown_text(self.entry))
        self.field.set_client(self.record, value)

    def display(self):
        self.update_selection(self.record, self.field)
        self.set_popdown(self.selection, self.entry)
        if not self.field:
            self.entry.set_active(-1)
            # When setting no item GTK doesn't clear the entry
            self.entry.get_child().set_text('')
            return
        super(Selection, self).display()
        value = self.field.get(self.record)
        if isinstance(value, (list, tuple)):
            # Compatibility with Many2One
            value = value[0]

        self.entry.handler_block_by_func(self.changed)
        if not self.set_popdown_value(self.entry, value):
            text = self.get_inactive_selection(value)
            self.set_popdown(self.selection[:] + [(value, text)], self.entry)
            self.set_popdown_value(self.entry, value)
        self.entry.handler_unblock_by_func(self.changed)



# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.

from gi.repository import Gtk


class InfoBar(object):

    toolbar = None

    def create_info_bar(self):
        self.__box = Gtk.VBox()
        self.__box.show()
        self.__messages = set()
        self.__kinds = {}
        return self.__box

    def info_bar_add(self, message, type_=Gtk.MessageType.ERROR, kind=None):
        if not message:
            return
        key = (message, type_)
        if key not in self.__messages:
            info_bar = Gtk.InfoBar()
            self.__box.pack_start(info_bar, False, False, 0)
            area = info_bar.get_content_area()
            area.add(Gtk.Label(label=message))
            info_bar.set_show_close_button(True)
            info_bar.connect('response', self.__response, key)
            info_bar.set_message_type(type_)
            info_bar.show_all()
            self.__kinds[info_bar] = kind

    def __response(self, widget, response, key):
        self.__messages.add(key)
        self.__box.remove(widget)

    def info_bar_refresh(self, kind=None):
        for child in self.__box.get_children():
            if self.__kinds[child] == kind:
                self.__box.remove(child)
                del self.__kinds[child]

    def info_bar_clear(self):
        for kind in set(self.__kinds.values()):
            self.info_bar_refresh(kind=kind)
        self.__messages.clear()



# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
import gettext

from gi.repository import Gdk, GLib, Gtk

import tryton.common as common
from tryton.common import TRYTON_ICON, RPCException, RPCExecute
from tryton.common.underline import set_underline
from tryton.common.widget_style import widget_class
from tryton.gui import Main
from tryton.gui.window.nomodal import NoModal

_ = gettext.gettext


class Widget(object):
    expand = False
    default_width_chars = 8

    def __init__(self, view, attrs):
        super(Widget, self).__init__()
        self.view = view
        self.attrs = attrs
        self.widget = None
        self.mnemonic_widget = None
        self.visible = True
        self._readonly = False

    @property
    def field_name(self):
        return self.attrs['name']

    @property
    def model_name(self):
        return self.view.screen.model_name

    @property
    def record(self):
        return self.view.record

    @property
    def field(self):
        if self.record:
            return self.record.group.fields[self.field_name]
        return None

    def destroy(self):
        pass

    def sig_activate(self, widget=None):
        # emulate a focus_out so that the onchange is called if needed
        self._focus_out()

    def _readonly_set(self, readonly):
        self._readonly = readonly

    def _required_set(self, required):
        pass

    def _invisible_widget(self):
        return self.widget

    @property
    def _invalid_widget(self):
        return self.widget

    @property
    def _required_widget(self):
        return self.widget

    @property
    def modified(self):
        return False

    def send_modified(self, *args):
        def send(value):
            if not self.widget.props.window:
                return
            if self.record and self.get_value() == value and self.modified:
                self.view.screen.record_modified(display=False)

        def get_value():
            if not self.widget.props.window:
                return
            GLib.timeout_add(300, send, self.get_value())
        # Wait the current event is finished to retreive the value
        GLib.idle_add(get_value)
        return False

    def invisible_set(self, value):
        widget = self._invisible_widget()
        if value and value != '0':
            self.visible = False
            widget.hide()
        else:
            self.visible = True
            widget.show()

    def _focus_out(self, *args):
        if not self.field:
            return False
        if not self.visible:
            return False
        self.set_value()

    def display(self):
        if not self.field:
            self._readonly_set(self.attrs.get('readonly', True))
            self.invisible_set(self.attrs.get('invisible', False))
            self._required_set(False)
            return
        states = self.field.get_state_attrs(self.record)
        readonly = self.attrs.get('readonly', states.get('readonly', False))
        if self.view.screen.readonly:
            readonly = True
        self._readonly_set(readonly)
        widget_class(self.widget, 'readonly', readonly)
        self._required_set(not readonly and states.get('required', False))
        widget_class(
            self._required_widget, 'required',
            not readonly and states.get('required', False))
        invalid = states.get('invalid', False)
        widget_class(self._invalid_widget, 'invalid', not readonly and invalid)
        self.invisible_set(self.attrs.get(
                'invisible', states.get('invisible', False)))

    def get_value(self):
        pass

    def set_value(self):
        pass


class TranslateDialog(NoModal):

    def __init__(self, widget, languages, readonly):
        NoModal.__init__(self)
        self.widget = widget
        self.win = Gtk.Dialog(
            title=_('Translation'), transient_for=self.parent,
            destroy_with_parent=True)
        Main().add_window(self.win)
        self.win.set_position(Gtk.WindowPosition.CENTER_ON_PARENT)
        self.win.set_icon(TRYTON_ICON)
        self.win.connect('response', self.response)
        self.win.set_default_size(*self.default_size())

        self.accel_group = Gtk.AccelGroup()
        self.win.add_accel_group(self.accel_group)

        cancel_button = self.win.add_button(
            set_underline(_("Cancel")), Gtk.ResponseType.CANCEL)
        cancel_button.set_image(
            common.IconFactory.get_image('tryton-cancel', Gtk.IconSize.BUTTON))
        cancel_button.set_always_show_image(True)
        ok_button = self.win.add_button(
            set_underline(_("OK")), Gtk.ResponseType.OK)
        ok_button.set_image(
            common.IconFactory.get_image('tryton-ok', Gtk.IconSize.BUTTON))
        ok_button.set_always_show_image(True)
        ok_button.add_accelerator(
            'clicked', self.accel_group, Gdk.KEY_Return,
            Gdk.ModifierType.CONTROL_MASK, Gtk.AccelFlags.VISIBLE)

        tooltips = common.Tooltips()

        self.widgets = {}
        grid = Gtk.Grid(column_spacing=3, row_spacing=3)
        for i, language in enumerate(languages):
            label = language['name'] + _(':')
            label = Gtk.Label(
                label=label,
                halign=Gtk.Align.END,
                valign=(Gtk.Align.START if self.widget.expand
                    else Gtk.Align.FILL))
            grid.attach(label, 0, i, 1, 1)

            context = dict(
                language=language['code'],
                fuzzy_translation=False,
                )
            try:
                value = RPCExecute('model', self.widget.record.model_name,
                    'read', [self.widget.record.id], [self.widget.field_name],
                    context={'language': language['code']}
                    )[0][self.widget.field_name]
            except RPCException:
                return
            context['fuzzy_translation'] = True
            try:
                fuzzy_value = RPCExecute('model',
                    self.widget.record.model_name, 'read',
                    [self.widget.record.id], [self.widget.field_name],
                    context=context)[0][self.widget.field_name]
            except RPCException:
                return
            if fuzzy_value is None:
                fuzzy_value = ''
            widget = self.widget.translate_widget()
            label.set_mnemonic_widget(widget)
            self.widget.translate_widget_set(widget, fuzzy_value)
            self.widget.translate_widget_set_readonly(widget, True)
            widget.set_vexpand(self.widget.expand)
            widget.set_hexpand(True)
            grid.attach(widget, 1, i, 1, 1)
            editing = Gtk.ToggleButton(label=_("Edit"))
            editing.connect('toggled', self.editing_toggled, widget)
            editing.props.sensitive = not readonly
            grid.attach(editing, 2, i, 1, 1)
            if value != fuzzy_value:
                fuzzy = Gtk.Label(_("Fuzzy"))
                widget_class(fuzzy, 'warning', True)
                grid.attach(fuzzy, 4, i, 1, 1)
            self.widgets[language['code']] = (widget, editing)

        tooltips.enable()
        vbox = Gtk.VBox()
        vbox.pack_start(grid, expand=self.widget.expand, fill=True, padding=0)
        viewport = Gtk.Viewport()
        viewport.set_shadow_type(Gtk.ShadowType.NONE)
        viewport.add(vbox)
        scrolledwindow = Gtk.ScrolledWindow()
        scrolledwindow.set_policy(
            Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        scrolledwindow.set_shadow_type(Gtk.ShadowType.NONE)
        scrolledwindow.add(viewport)
        self.win.vbox.pack_start(
            scrolledwindow, expand=True, fill=True, padding=0)
        self.win.show_all()

        self.register()
        self.show()

    def editing_toggled(self, editing, widget):
        self.widget.translate_widget_set_readonly(widget,
            not editing.get_active())

    def response(self, win, response):
        if response == Gtk.ResponseType.OK:
            for code, widget in self.widgets.items():
                widget, editing = widget
                if not editing.get_active():
                    continue
                value = self.widget.translate_widget_get(widget)
                context = dict(
                    language=code,
                    fuzzy_translation=False,
                    )
                try:
                    RPCExecute('model', self.widget.record.model_name, 'write',
                        [self.widget.record.id], {
                            self.widget.field_name: value,
                            }, context=context)
                except RPCException:
                    pass
            self.widget.record.cancel()
            self.widget.view.display()
        self.destroy()

    def destroy(self):
        self.win.destroy()
        NoModal.destroy(self)

    def show(self):
        self.win.show()

    def hide(self):
        self.win.hide()


class TranslateMixin:

    def translate_button(self):
        button = Gtk.Button()
        button.set_image(common.IconFactory.get_image(
                'tryton-translate', Gtk.IconSize.SMALL_TOOLBAR))
        button.set_relief(Gtk.ReliefStyle.NONE)
        button.connect('clicked', self.translate)
        return button

    def translate(self, widget=None, *args):
        if widget:
            widget.grab_focus()
        self.view.set_value()
        if self.record.id < 0 or self.record.modified:
            common.message(
                _('You need to save the record before adding translations.'))
            return

        try:
            lang_ids = RPCExecute('model', 'ir.lang', 'search', [
                    ('translatable', '=', True),
                    ])
        except RPCException:
            return

        if not lang_ids:
            common.message(_('No other language available.'))
            return
        try:
            languages = RPCExecute('model', 'ir.lang', 'read', lang_ids,
                ['code', 'name'])
        except RPCException:
            return

        self.translate_dialog(languages)

    def translate_dialog(self, languages):
        TranslateDialog(self, languages, self._readonly)

    def translate_widget(self):
        raise NotImplementedError

    def translate_widget_set(self, widget, value):
        raise NotImplementedError

    def translate_widget_get(self, widget):
        raise NotImplementedError

    def translate_widget_set_readonly(self, widget, value):
        raise NotImplementedError



# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
'Action'
import gettext

from gi.repository import GLib, Gtk

import tryton.rpc as rpc
from tryton.action import Action as GenericAction
from tryton.common import RPCException, RPCExecute
from tryton.config import CONFIG
from tryton.gui.window.view_form.screen import Screen
from tryton.gui.window.win_form import WinForm
from tryton.pyson import PYSONDecoder

_ = gettext.gettext


class Action:

    def __init__(self, view, attrs=None):
        super(Action, self).__init__()
        self.name = attrs['name']
        self.view = view

        try:
            self.action = RPCExecute('model', 'ir.action.act_window', 'get',
                self.name)
        except RPCException:
            raise

        view_ids = []
        view_mode = None
        if self.action.get('views', []):
            view_ids = [x[0] for x in self.action['views']]
            view_mode = [x[1] for x in self.action['views']]
        elif self.action.get('view_id', False):
            view_ids = [self.action['view_id'][0]]

        self.action.setdefault('pyson_domain', '[]')
        ctx = {}
        ctx.update(rpc.CONTEXT)
        ctx['_user'] = rpc._USER
        decoder = PYSONDecoder(ctx)
        action_ctx = self.view.context.copy()
        action_ctx.update(
            decoder.decode(self.action.get('pyson_context') or '{}'))
        ctx.update(action_ctx)

        ctx['context'] = ctx
        decoder = PYSONDecoder(ctx)
        order = decoder.decode(self.action['pyson_order'])
        search_value = decoder.decode(
            self.action['pyson_search_value'] or '[]')
        tab_domain = [(n, decoder.decode(d), c)
            for n, d, c in self.action['domains']]

        limit = self.action.get('limit')
        if limit is None:
            limit = CONFIG['client.limit']

        self.context = ctx
        self.domain = []
        self.update_domain([])

        self.widget = Gtk.Frame()
        self.widget.set_border_width(0)

        vbox = Gtk.VBox(homogeneous=False, spacing=3)
        self.widget.add(vbox)

        self.title = Gtk.Label()
        self.widget.set_label_widget(self.title)
        self.widget.set_label_align(0.0, 0.5)
        self.widget.show_all()

        self.screen = Screen(self.action['res_model'],
            view_ids=view_ids,
            domain=self.domain,
            context=action_ctx,
            order=order,
            mode=view_mode,
            limit=limit,
            search_value=search_value,
            tab_domain=tab_domain,
            context_model=self.action['context_model'],
            context_domain=self.action['context_domain'],
            row_activate=self.row_activate)
        self.screen.windows.append(self)
        vbox.pack_start(
            self.screen.widget, expand=True, fill=True, padding=0)

        if attrs.get('string'):
            self.title.set_text(attrs['string'])
        else:
            self.title.set_text(self.action['name'])

        self.widget.set_size_request(int(attrs.get('width', -1)),
                int(attrs.get('height', -1)))

        self.screen.search_filter()

    def row_activate(self):
        if not self.screen.current_record:
            return

        if (self.screen.current_view.view_type == 'tree'
                and int(self.screen.current_view.attributes.get(
                        'keyword_open', 0))):
            GenericAction.exec_keyword('tree_open', {
                    'model': self.screen.model_name,
                    'id': (self.screen.current_record.id
                        if self.screen.current_record else None),
                    'ids': [r.id for r in self.screen.selected_records],
                    }, context=self.screen.group._context.copy(),
                warning=False)
        else:
            def callback(result):
                if result:
                    self.screen.current_record.save()
                else:
                    self.screen.current_record.cancel()
            WinForm(self.screen, callback, title=self.title.get_text())

    def display(self):
        self.screen.search_filter(self.screen.screen_container.get_text())

    def record_message(self, *args):
        self.view.active_changed(self)

    def _get_active(self):
        if self.screen and self.screen.current_record:
            return {
                'active_id': self.screen.current_record.id,
                'active_ids': [r.id for r in self.screen.selected_records]
                }
        else:
            return {
                'active_id': None,
                'active_ids': [],
                }

    active = property(_get_active)

    def update_domain(self, actions):
        domain_ctx = self.context.copy()
        domain_ctx['_actions'] = {}
        for action in actions:
            domain_ctx['_actions'][action.name] = action.active
        new_domain = PYSONDecoder(domain_ctx).decode(
                self.action['pyson_domain'])
        if self.domain == new_domain:
            return
        del self.domain[:]
        self.domain.extend(new_domain)
        if hasattr(self, 'screen'):  # Catch early update
            # Using idle_add to prevent corruption of the event who triggered
            # the update.
            def display():
                if self.screen.widget.props.window:
                    self.display()
            GLib.idle_add(display)



# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
import datetime
import decimal
import locale
import logging
import math
import os
import tempfile
from decimal import Decimal
from itertools import chain
from pathlib import Path

import tryton.common as common
from tryton.common import (
    EvalEnvironment, RPCException, RPCExecute, concat, domain_inversion,
    eval_domain, extract_reference_models, filter_leaf, inverse_leaf,
    localize_domain, merge, prepare_reference_domain, simplify, unique_value)
from tryton.common.htmltextbuffer import guess_decode
from tryton.config import CONFIG
from tryton.pyson import PYSONDecoder

logger = logging.getLogger(__name__)


class Field(object):
    '''
    get: return the values to write to the server
    get_client: return the value for the client widget (form_gtk)
    set: save the value from the server
    set_client: save the value from the widget
    '''
    _default = None
    _single_value = True

    @staticmethod
    def get_field(ctype):
        return TYPES.get(ctype, CharField)

    def __init__(self, attrs):
        self.attrs = attrs
        self.name = attrs['name']
        self.views = set()

    def sig_changed(self, record):
        record.on_change([self.name])
        record.on_change_with([self.name])
        record.autocomplete_with(self.name)
        record.set_field_context()

    def domains_get(self, record, pre_validate=None):
        screen_domain = domain_inversion(
            [record.group.domain4inversion, pre_validate or []],
            self.name, EvalEnvironment(record))
        if isinstance(screen_domain, bool) and not screen_domain:
            screen_domain = [('id', '=', None)]
        elif isinstance(screen_domain, bool) and screen_domain:
            screen_domain = []
        attr_domain = record.expr_eval(self.attrs.get('domain', []))
        return screen_domain, attr_domain

    def domain_get(self, record):
        screen_domain, attr_domain = self.domains_get(record)
        return concat(localize_domain(screen_domain), attr_domain)

    def validation_domains(self, record, pre_validate=None):
        return concat(*self.domains_get(record, pre_validate))

    def get_context(self, record, record_context=None, local=False):
        if record_context is not None:
            context = record_context.copy()
        else:
            context = record.get_context(local=local)
        context.update(record.expr_eval(self.attrs.get('context', {})))
        return context

    def get_search_context(self, record):
        context = self.get_context(record)
        context.update(record.expr_eval(self.attrs.get('search_context', {})))
        return context

    def get_search_order(self, record):
        order = record.expr_eval(self.attrs.get('search_order', None))
        return order

    def _is_empty(self, record):
        return not self.get_eval(record)

    def check_required(self, record):
        state_attrs = self.get_state_attrs(record)
        if bool(int(state_attrs.get('required') or 0)):
            if (self._is_empty(record)
                    and not bool(int(state_attrs.get('readonly') or 0))):
                return False
        return True

    def validate(self, record, softvalidation=False, pre_validate=None):
        if self.attrs.get('readonly'):
            return True
        state_attrs = self.get_state_attrs(record)
        is_required = bool(int(state_attrs.get('required') or 0))
        is_invisible = bool(int(state_attrs.get('invisible') or 0))
        invalid = False
        state_attrs['domain_readonly'] = False
        domain = simplify(self.validation_domains(record, pre_validate))
        if not softvalidation:
            if not self.check_required(record):
                invalid = 'required'
        if isinstance(domain, bool):
            if not domain:
                invalid = 'domain'
        elif domain == [('id', '=', None)]:
            invalid = 'domain'
        else:
            screen_domain, _ = self.domains_get(record, pre_validate)
            unique, leftpart, value = unique_value(
                domain, single_value=self._single_value)
            unique_from_screen, _, _ = unique_value(
                screen_domain, single_value=self._single_value)
            if (self._is_empty(record)
                    and not is_required
                    and not is_invisible
                    and not unique_from_screen):
                pass
            elif unique:
                # If the inverted domain is so constraint that only one value
                # is possible we should use it. But we must also pay attention
                # to the fact that the original domain might be a 'OR' domain
                # and thus not preventing the modification of fields.
                if value is False:
                    # XXX to remove once server domains are fixed
                    value = None
                if record.group.domain:
                    original_domain = merge(record.group.domain)
                else:
                    original_domain = merge(domain)
                domain_readonly = original_domain[0] == 'AND'
                if '.' in leftpart:
                    localpart = leftpart.split('.', 1)[1]
                    setdefault = localpart == 'id'
                else:
                    setdefault = True
                if setdefault and not pre_validate:
                    self.set_client(record, value)
                    state_attrs['domain_readonly'] = domain_readonly
            if not eval_domain(domain, EvalEnvironment(record)):
                invalid = domain
        state_attrs['invalid'] = invalid
        return not invalid

    def set(self, record, value):
        record.value[self.name] = value

    def get(self, record):
        return record.value.get(self.name, self._default)

    def get_eval(self, record):
        return self.get(record)

    def get_on_change_value(self, record):
        return self.get_eval(record)

    def set_client(self, record, value, force_change=False):
        previous_value = self.get(record)
        self.set(record, value)
        if previous_value != self.get(record):
            self.sig_changed(record)
            record.validate(softvalidation=True)
            record.set_modified(self.name)
        elif force_change:
            self.sig_changed(record)
            record.validate(softvalidation=True)
            record.set_modified()

    def get_client(self, record):
        return self.get(record)

    def set_default(self, record, value):
        self.set(record, value)
        record.modified_fields.setdefault(self.name)

    def set_on_change(self, record, value):
        record.modified_fields.setdefault(self.name)
        self.set(record, value)

    def state_set(self, record, states=('readonly', 'required', 'invisible')):
        state_changes = record.expr_eval(self.attrs.get('states', {}))
        for key in states:
            if key == 'readonly' and self.attrs.get(key, False):
                continue
            if key in state_changes:
                self.get_state_attrs(record)[key] = state_changes[key]
            elif key in self.attrs:
                self.get_state_attrs(record)[key] = self.attrs[key]
        if (record.group.readonly
                or self.get_state_attrs(record).get('domain_readonly')
                or record.parent_name == self.name):
            self.get_state_attrs(record)['readonly'] = True

    def get_state_attrs(self, record):
        if self.name not in record.state_attrs:
            record.state_attrs[self.name] = self.attrs.copy()
        if record.group.readonly or record.readonly:
            record.state_attrs[self.name]['readonly'] = True
        return record.state_attrs[self.name]

    def get_timestamp(self, record):
        return {}


class CharField(Field):
    _default = ''

    def set(self, record, value):
        if self.attrs.get('strip') and value:
            if self.attrs['strip'] == 'leading':
                value = value.lstrip()
            elif self.attrs['strip'] == 'trailing':
                value = value.rstrip()
            else:
                value = value.strip()
        super().set(record, value)

    def get(self, record):
        return super(CharField, self).get(record) or self._default

    def set_client(self, record, value, force_change=False):
        if isinstance(value, bytes):
            try:
                value = guess_decode(value)
            except UnicodeDecodeError:
                logger.warning(
                    "The encoding can not be guessed for field '%(name)s'",
                    {'name': self.name})
                value = None
        super().set_client(record, value, force_change)


class SelectionField(Field):

    _default = None

    def set_client(self, record, value, force_change=False):
        record.value.pop(self.name + ':string', None)
        super().set_client(record, value, force_change=force_change)


class MultiSelectionField(SelectionField):

    _default = None
    _single_value = False

    def get(self, record):
        value = super().get(record)
        if not value:
            value = self._default
        else:
            value.sort()
        return value

    def get_eval(self, record):
        value = super().get_eval(record)
        if value is None:
            value = []
        return value

    def set_client(self, record, value, force_change=False):
        if value is None:
            value = []
        if isinstance(value, str):
            value = [value]
        if value:
            value = sorted(value)
        super().set_client(record, value, force_change=force_change)


class DateTimeField(Field):

    _default = None

    def set_client(self, record, value, force_change=False):
        if isinstance(value, datetime.time):
            current_value = self.get_client(record)
            if current_value:
                value = datetime.datetime.combine(
                    current_value.date(), value)
            else:
                value = None
        elif value and not isinstance(value, datetime.datetime):
            current_value = self.get_client(record)
            if current_value:
                time = current_value.time()
            else:
                time = datetime.time()
            value = datetime.datetime.combine(value, time)
        if value:
            value = common.untimezoned_date(value)
        super(DateTimeField, self).set_client(record, value,
            force_change=force_change)

    def get_client(self, record):
        value = super(DateTimeField, self).get_client(record)
        if value:
            return common.timezoned_date(value)

    def date_format(self, record):
        context = self.get_context(record)
        return common.date_format(context.get('date_format'))

    def time_format(self, record):
        return record.expr_eval(self.attrs['format'])


class DateField(Field):

    _default = None

    def set_client(self, record, value, force_change=False):
        if isinstance(value, datetime.datetime):
            assert value.time() == datetime.time()
            value = value.date()
        super(DateField, self).set_client(record, value,
            force_change=force_change)

    def date_format(self, record):
        context = self.get_context(record)
        return common.date_format(context.get('date_format'))


class TimeField(Field):

    _default = None

    def _is_empty(self, record):
        return self.get(record) is None

    def set_client(self, record, value, force_change=False):
        if isinstance(value, datetime.datetime):
            value = value.time()
        super(TimeField, self).set_client(record, value,
            force_change=force_change)

    def time_format(self, record):
        return record.expr_eval(self.attrs['format'])


class TimeDeltaField(Field):

    _default = None

    def _is_empty(self, record):
        return self.get(record) is None

    def converter(self, group):
        return group.context.get(self.attrs.get('converter'))

    def set_client(self, record, value, force_change=False):
        if isinstance(value, str):
            value = common.timedelta.parse(value, self.converter(record.group))
        super(TimeDeltaField, self).set_client(
            record, value, force_change=force_change)

    def get_client(self, record):
        value = super(TimeDeltaField, self).get_client(record)
        return common.timedelta.format(value, self.converter(record.group))


class FloatField(Field):
    _default = None

    def __init__(self, attrs):
        super().__init__(attrs)
        self._digits = {}
        self._symbol = {}

    def _is_empty(self, record):
        return self.get(record) is None

    def get(self, record):
        return record.value.get(self.name, self._default)

    def digits(self, record, factor=1):
        digits = record.expr_eval(self.attrs.get('digits'))
        if isinstance(digits, str):
            if digits not in record.group.fields:
                return
            digits_field = record.group.fields[digits]
            digits_name = digits_field.attrs.get('relation')
            digits_id = digits_field.get(record)
            if digits_name and digits_id is not None and digits_id >= 0:
                if digits_id in self._digits:
                    digits = self._digits[digits_id]
                else:
                    try:
                        digits = RPCExecute(
                            'model', digits_name, 'get_digits', digits_id)
                    except RPCException:
                        logger.warn(
                            "Fail to fetch digits for %s,%s",
                            digits_name, digits_id)
                        return
                    self._digits[digits_id] = digits
            else:
                return
        if not digits or any(d is None for d in digits):
            return
        shift = int(round(math.log(abs(factor), 10)))
        return (digits[0] + shift, digits[1] - shift)

    def get_symbol(self, record, symbol):
        if record and symbol in record.group.fields:
            value = self.get(record) or 0
            sign = 1
            if value < 0:
                sign = -1
            elif value == 0:
                sign = 0
            symbol_field = record.group.fields[symbol]
            symbol_name = symbol_field.attrs.get('relation')
            symbol_id = symbol_field.get(record)
            if symbol_name and symbol_id is not None and symbol_id >= 0:
                if symbol_id in self._symbol:
                    return self._symbol[symbol_id]
                try:
                    result = RPCExecute(
                        'model', symbol_name, 'get_symbol', symbol_id, sign,
                        context=record.get_context())
                    self._symbol[symbol_id] = result
                    return result
                except RPCException:
                    logger.warn(
                        "Fail to fetch symbol for %s,%s",
                        symbol_name, symbol_id)
        return '', 1

    def convert(self, value):
        if value is None:
            return None
        if isinstance(value, str):
            value = locale.delocalize(
                value, self.attrs.get('monetary', False))
        try:
            return self._convert(value)
        except ValueError:
            return self._default
    _convert = float

    def apply_factor(self, record, value, factor):
        if value is not None:
            # The default precision is the one used by value (before
            # applying the factor), as python uses the shortest representation
            # of said value.
            # Once the factor is applied the number might become even
            # more inexact thus we should rely on the initial
            # precision + the effect factor will have
            default_precision = len((str(value).split('.', 2) + [''])[1])
            default_precision += math.ceil(math.log10(factor))
            value /= factor
            digits = self.digits(record)
            if digits:
                value = round(value, digits[1])
            else:
                value = round(value, default_precision)
            value = self.convert(value)
        return value

    def set_client(self, record, value, force_change=False, factor=1):
        value = self.apply_factor(record, self.convert(value), factor)
        super(FloatField, self).set_client(record, value,
            force_change=force_change)

    def get_client(self, record, factor=1, grouping=True):
        value = record.value.get(self.name)
        if value is not None:
            digits = self.digits(record, factor=factor)
            d = value * factor
            if not isinstance(d, Decimal):
                d = Decimal(repr(d))
            if digits:
                p = int(digits[1])
            elif d == d.to_integral_value():
                p = 0
            else:
                p = -int(d.as_tuple().exponent)
            monetary = self.attrs.get('monetary', False)
            return locale.localize(
                '{0:.{1}f}'.format(d, p), grouping, monetary)
        else:
            return ''


class NumericField(FloatField):

    def convert(self, value):
        try:
            return super().convert(value)
        except decimal.InvalidOperation:
            return self._default
    _convert = Decimal

    def set_client(self, record, value, force_change=False, factor=1):
        return super(NumericField, self).set_client(record, value,
            force_change=force_change, factor=Decimal(str(factor)))

    def get_client(self, record, factor=1, grouping=True):
        return super(NumericField, self).get_client(record,
            factor=Decimal(str(factor)), grouping=grouping)


class IntegerField(FloatField):

    _convert = int

    def set_client(self, record, value, force_change=False, factor=1):
        return super(IntegerField, self).set_client(record, value,
            force_change=force_change, factor=int(factor))

    def get_client(self, record, factor=1, grouping=True):
        return super(IntegerField, self).get_client(
            record, factor=int(factor), grouping=grouping)


class BooleanField(Field):

    _default = False

    def set_client(self, record, value, force_change=False):
        value = bool(value)
        super(BooleanField, self).set_client(record, value,
            force_change=force_change)

    def get(self, record):
        return bool(record.value.get(self.name))

    def get_client(self, record):
        return bool(record.value.get(self.name))


class M2OField(Field):
    '''
    internal = (id, name)
    '''

    _default = None

    def _is_empty(self, record):
        return self.get(record) is None

    def get_client(self, record):
        rec_name = record.value.get(self.name + '.', {}).get('rec_name')
        if rec_name is None:
            self.set(record, self.get(record))
            rec_name = record.value.get(self.name + '.', {}).get('rec_name')
        return rec_name or ''

    def set_client(self, record, value, force_change=False):
        if isinstance(value, (tuple, list)):
            value, rec_name = value
        else:
            if value == self.get(record):
                rec_name = record.value.get(
                    self.name + '.', {}).get('rec_name', '')
            else:
                rec_name = ''
        if value and value < 0 and self.name != record.parent_name:
            value, rec_name = None, ''
        record.value.setdefault(self.name + '.', {})['rec_name'] = rec_name
        super(M2OField, self).set_client(record, value,
            force_change=force_change)

    def set(self, record, value):
        rec_name = record.value.get(self.name + '.', {}).get('rec_name') or ''
        if not rec_name and value is not None and value >= 0:
            try:
                result, = RPCExecute('model', self.attrs['relation'], 'read',
                    [value], ['rec_name'])
            except RPCException:
                return
            rec_name = result['rec_name'] or ''
        record.value.setdefault(self.name + '.', {})['rec_name'] = rec_name
        record.value[self.name] = value

    def get_context(self, record, record_context=None, local=False):
        context = super(M2OField, self).get_context(
            record, record_context=record_context, local=local)
        if self.attrs.get('datetime_field'):
            context['_datetime'] = record.get_eval(
                )[self.attrs.get('datetime_field')]
        return context

    def validation_domains(self, record, pre_validate=None):
        screen_domain, attr_domain = self.domains_get(record, pre_validate)
        return screen_domain

    def domain_get(self, record):
        screen_domain, attr_domain = self.domains_get(record)
        return concat(localize_domain(inverse_leaf(screen_domain), self.name),
            attr_domain)

    def get_on_change_value(self, record):
        if record.parent_name == self.name and record.parent:
            return record.parent.get_on_change_value(
                skip={record.group.child_name})
        return super(M2OField, self).get_on_change_value(record)


class O2OField(M2OField):
    pass


class O2MField(Field):
    '''
    internal = Group of the related objects
    '''

    _default = None
    _single_value = False

    def __init__(self, attrs):
        super(O2MField, self).__init__(attrs)

    def _set_default_value(self, record, fields=None):
        if record.value.get(self.name) is not None:
            return
        from .group import Group
        parent_name = self.attrs.get('relation_field', '')
        fields = fields or {}
        group = Group(self.attrs['relation'], fields,
                parent=record,
                parent_name=parent_name,
                child_name=self.name,
                parent_datetime_field=self.attrs.get('datetime_field'))
        if not fields and record.model_name == self.attrs['relation']:
            group.fields = record.group.fields
        record.value[self.name] = group

    def get_client(self, record):
        self._set_default_value(record)
        return record.value.get(self.name)

    def get(self, record):
        if record.value.get(self.name) is None:
            return []
        record_removed = record.value[self.name].record_removed
        record_deleted = record.value[self.name].record_deleted
        result = []
        parent_name = self.attrs.get('relation_field', '')
        to_add = []
        to_create = []
        to_write = []
        for record2 in record.value[self.name]:
            if record2 in record_removed or record2 in record_deleted:
                continue
            if record2.id >= 0:
                if record2.modified:
                    values = record2.get()
                    values.pop(parent_name, None)
                    if values:
                        to_write.extend(([record2.id], values))
                    to_add.append(record2.id)
            else:
                values = record2.get()
                values.pop(parent_name, None)
                to_create.append(values)
        if to_add:
            result.append(('add', to_add))
        if to_create:
            result.append(('create', to_create))
        if to_write:
            result.append(('write',) + tuple(to_write))
        if record_removed:
            result.append(('remove', [x.id for x in record_removed]))
        if record_deleted:
            result.append(('delete', [x.id for x in record_deleted]))
        return result

    def get_timestamp(self, record):
        if record.value.get(self.name) is None:
            return {}
        result = {}
        record_modified = (r for r in record.value[self.name] if r.modified)
        for record2 in chain(record_modified,
                record.value[self.name].record_removed,
                record.value[self.name].record_deleted):
            result.update(record2.get_timestamp())
        return result

    def get_eval(self, record):
        if record.value.get(self.name) is None:
            return []
        record_removed = record.value[self.name].record_removed
        record_deleted = record.value[self.name].record_deleted
        return [x.id for x in record.value[self.name]
            if x not in record_removed and x not in record_deleted]

    def get_on_change_value(self, record):
        result = []
        if record.value.get(self.name) is None:
            return []
        for record2 in record.value[self.name]:
            if not (record2.deleted or record2.removed):
                result.append(
                    record2.get_on_change_value(
                        skip={self.attrs.get('relation_field', '')}))
        return result

    def _set_value(self, record, value, default=False, modified=False):
        self._set_default_value(record)
        group = record.value[self.name]
        if value is None:
            value = []
        if not value or isinstance(value[0], int):
            mode = 'list ids'
        else:
            mode = 'list values'

        if mode == 'list values':
            context = self.get_context(record)
            field_names = set(f for v in value for f in v
                if f not in group.fields and '.' not in f)
            if field_names:
                try:
                    fields = RPCExecute('model', self.attrs['relation'],
                        'fields_get', list(field_names), context=context)
                except RPCException:
                    return
                group.load_fields(fields)

        if mode == 'list ids':
            records_to_remove = [r for r in group if r.id not in value]
            for record_to_remove in records_to_remove:
                group.remove(record_to_remove, remove=True, modified=False)
            group.load(value, modified=modified or default)
        else:
            for vals in value:
                if 'id' in vals:
                    new_record = group.get(vals['id'])
                    if not new_record:
                        new_record = group.new(
                            default=False, obj_id=vals['id'])
                else:
                    new_record = group.new(default=False)
                if default:
                    # Don't validate as parent will validate
                    new_record.set_default(
                        vals, modified=False, validate=False)
                    group.add(new_record, modified=False)
                else:
                    new_record.set(vals, modified=False)
                    group.append(new_record)
            # Trigger modified only once
            group.record_modified()

    def set(self, record, value, _default=False):
        group = record.value.get(self.name)
        fields = {}
        if group is not None:
            fields = group.fields.copy()
            # Unconnect to prevent infinite loop
            group.parent = None
            group.destroy()
        elif record.model_name == self.attrs['relation']:
            fields = record.group.fields
        if fields:
            fields = dict((fname, field.attrs)
                for fname, field in fields.items())

        record.value[self.name] = None
        self._set_default_value(record, fields=fields)
        group = record.value[self.name]

        # Prevent to trigger group-cleared
        group.parent = None
        self._set_value(record, value, default=_default)
        group.parent = record

    def set_client(self, record, value, force_change=False):
        # domain inversion could try to set None as value
        if value is None:
            value = []
        # domain inversion could try to set id as value
        if isinstance(value, int):
            value = [value]

        previous_ids = self.get_eval(record)
        # The order of the ids is not significant
        modified = set(previous_ids) != set(value)
        self._set_value(record, value, modified=modified)
        if modified:
            self.sig_changed(record)
            record.validate(softvalidation=True)
            record.set_modified(self.name)
        elif force_change:
            self.sig_changed(record)
            record.validate(softvalidation=True)
            record.set_modified()

    def set_default(self, record, value):
        self.set(record, value, _default=True)
        record.modified_fields.setdefault(self.name)

    def set_on_change(self, record, value):
        record.modified_fields.setdefault(self.name)
        self._set_default_value(record)
        if isinstance(value, (list, tuple)):
            self._set_value(record, value, modified=True)
            return

        if value and (value.get('add') or value.get('update')):
            context = self.get_context(record)
            fields = record.value[self.name].fields
            values = chain(value.get('update', []),
                (d for _, d in value.get('add', [])))
            field_names = set(f for v in values
                for f in v if f not in fields and f != 'id' and '.' not in f)
            if field_names:
                try:
                    new_fields = RPCExecute('model', self.attrs['relation'],
                        'fields_get', list(field_names), context=context)
                except RPCException:
                    return
            else:
                new_fields = {}

        group = record.value[self.name]
        if value and value.get('delete'):
            for record_id in value['delete']:
                record2 = group.get(record_id)
                if record2 is not None:
                    group.remove(
                        record2, remove=False, modified=False,
                        force_remove=False)
        if value and value.get('remove'):
            for record_id in value['remove']:
                record2 = group.get(record_id)
                if record2 is not None:
                    group.remove(
                        record2, remove=True, modified=False,
                        force_remove=False)

        if value and (value.get('add') or value.get('update', [])):
            vals_to_set = {}
            # First set already added fields to prevent triggering a
            # second on_change call
            for vals in value.get('update', []):
                if 'id' not in vals:
                    continue
                record2 = group.get(vals['id'])
                if record2 is not None:
                    vals_to_set = {
                        k: v for k, v in vals.items() if k not in new_fields}
                    record2.set_on_change(vals_to_set)

            record.value[self.name].add_fields(new_fields)
            for index, vals in value.get('add', []):
                new_record = None
                id_ = vals.pop('id', None)
                if id_ is not None:
                    new_record = group.get(id_)
                if not new_record:
                    new_record = group.new(obj_id=id_, default=False)
                group.add(new_record, index, modified=False)
                new_record.set_on_change(vals)

            for vals in value.get('update', []):
                if 'id' not in vals:
                    continue
                record2 = group.get(vals['id'])
                if record2 is not None:
                    to_set = {
                        k: v for k, v in vals.items()
                        if k not in vals_to_set}
                    record2.set_on_change(to_set)

    def validation_domains(self, record, pre_validate=None):
        screen_domain, attr_domain = self.domains_get(record, pre_validate)
        return screen_domain

    def validate(self, record, softvalidation=False, pre_validate=None):
        invalid = False
        ldomain = localize_domain(domain_inversion(
                record.group.clean4inversion(pre_validate or []), self.name,
                EvalEnvironment(record)), self.name)
        if isinstance(ldomain, bool):
            if ldomain:
                ldomain = []
            else:
                ldomain = [('id', '=', None)]
        for record2 in record.value.get(self.name, []):
            if not record2.loaded and record2.id >= 0 and not pre_validate:
                continue
            if not record2.validate(softvalidation=softvalidation,
                    pre_validate=ldomain):
                invalid = 'children'
        test = super(O2MField, self).validate(record, softvalidation,
            pre_validate)
        if test and invalid:
            self.get_state_attrs(record)['invalid'] = invalid
            return False
        return test

    def state_set(self, record, states=('readonly', 'required', 'invisible')):
        self._set_default_value(record)
        super(O2MField, self).state_set(record, states=states)

    def domain_get(self, record):
        screen_domain, attr_domain = self.domains_get(record)
        # Forget screen_domain because it only means at least one record
        # and not all records
        return attr_domain


class M2MField(O2MField):

    def get_on_change_value(self, record):
        return self.get_eval(record)


class ReferenceField(Field):

    _default = None

    def _is_empty(self, record):
        result = super(ReferenceField, self)._is_empty(record)
        if not result and (record.value[self.name] is None
                or record.value[self.name][1] < 0):
            result = True
        return result

    def get_client(self, record):
        if record.value.get(self.name):
            model, _ = record.value[self.name]
            name = record.value.get(self.name + '.', {}).get('rec_name') or ''
            return model, name
        else:
            return None

    def get(self, record):
        if (record.value.get(self.name)
                and record.value[self.name][0]
                and record.value[self.name][1] is not None
                and record.value[self.name][1] >= -1):
            return ','.join(map(str, record.value[self.name]))
        return None

    def set_client(self, record, value, force_change=False):
        if value:
            if isinstance(value, str):
                value = value.split(',')
            ref_model, ref_id = value
            if isinstance(ref_id, (tuple, list)):
                ref_id, rec_name = ref_id
            else:
                if ref_id:
                    try:
                        ref_id = int(ref_id)
                    except ValueError:
                        pass
                if '%s,%s' % (ref_model, ref_id) == self.get(record):
                    rec_name = record.value.get(
                        self.name + '.', {}).get('rec_name') or ''
                else:
                    rec_name = ''
            record.value.setdefault(self.name + '.', {})['rec_name'] = rec_name
            value = (ref_model, ref_id)
        super(ReferenceField, self).set_client(record, value,
            force_change=force_change)

    def set(self, record, value):
        if not value:
            record.value[self.name] = self._default
            return
        if isinstance(value, str):
            ref_model, ref_id = value.split(',')
            if not ref_id:
                ref_id = None
            else:
                try:
                    ref_id = int(ref_id)
                except ValueError:
                    pass
        else:
            ref_model, ref_id = value
        rec_name = record.value.get(self.name + '.', {}).get('rec_name') or ''
        if ref_model and ref_id is not None and ref_id >= 0:
            if not rec_name and ref_id >= 0:
                try:
                    result, = RPCExecute('model', ref_model, 'read', [ref_id],
                        ['rec_name'])
                except RPCException:
                    return
                rec_name = result['rec_name']
        elif ref_model:
            rec_name = ''
        else:
            rec_name = str(ref_id) if ref_id is not None else ''
        record.value[self.name] = ref_model, ref_id
        record.value.setdefault(self.name + '.', {})['rec_name'] = rec_name

    def get_context(self, record, record_context=None, local=False):
        context = super(ReferenceField, self).get_context(
            record, record_context, local=local)
        if self.attrs.get('datetime_field'):
            context['_datetime'] = record.get_eval(
                )[self.attrs.get('datetime_field')]
        return context

    def get_on_change_value(self, record):
        if record.parent_name == self.name and record.parent:
            return record.parent.model_name, record.parent.get_on_change_value(
                skip={record.group.child_name})
        return super(ReferenceField, self).get_on_change_value(record)

    def validation_domains(self, record, pre_validate=None):
        screen_domain, attr_domain = self.domains_get(record, pre_validate)
        return screen_domain

    def domains_get(self, record, pre_validate=None):
        if record.value.get(self.name):
            model = record.value[self.name][0]
        else:
            model = None
        screen_domain, attr_domain = super().domains_get(
            record, pre_validate=pre_validate)
        attr_domain = attr_domain.get(model, [])
        return screen_domain, attr_domain

    def domain_get(self, record):
        if record.value.get(self.name):
            model = record.value[self.name][0]
        else:
            model = None
        screen_domain, attr_domain = self.domains_get(record)
        screen_domain = filter_leaf(screen_domain, self.name, model)
        screen_domain = prepare_reference_domain(screen_domain, self.name)
        return concat(localize_domain(
                screen_domain, self.name, strip_target=True), attr_domain)

    def get_search_order(self, record):
        order = super().get_search_order(record)
        if order is not None:
            if record.value.get(self.name):
                model = record.value[self.name][0]
            else:
                model = None
            order = order.get(model)
        return order

    def get_models(self, record):
        screen_domain, attr_domain = self.domains_get(record)
        screen_domain = prepare_reference_domain(screen_domain, self.name)
        return extract_reference_models(
            concat(screen_domain, attr_domain), self.name)


class _FileCache(object):
    def __init__(self, data=None):
        _, filename = tempfile.mkstemp(prefix='tryton_')
        self.path = Path(filename)
        self.suffixes = {}
        if data:
            with open(self.path, 'wb') as fp:
                fp.write(data)

    @property
    def data(self):
        with open(self.path, 'rb') as fp:
            return fp.read()

    def __del__(self):
        try:
            os.remove(self.path)
        except IOError:
            pass
        for path in self.suffixes.values():
            try:
                os.remove(path)
            except IOError:
                pass

    def with_suffix(self, suffix):
        if suffix in self.suffixes:
            return self.suffixes[suffix]
        _, filename = tempfile.mkstemp(prefix='tryton_', suffix=suffix)
        self.suffixes[suffix] = path = Path(filename)
        with open(path, 'wb') as fp:
            fp.write(self.data)
        return path


class BinaryField(Field):

    _default = None

    def _set_file_cache(self, record, data):
        if isinstance(data, str):
            data = data.encode('utf-8')
        file_cache = _FileCache(data)
        self.set(record, file_cache)
        return file_cache

    def get(self, record):
        result = record.value.get(self.name, self._default)
        if isinstance(result, _FileCache):
            try:
                result = result.data
            except IOError:
                result = self.get_data(record)
        return result

    def get_client(self, record):
        return self.get(record)

    def set_client(self, record, value, force_change=False):
        self._set_file_cache(record, value or b'')
        self.sig_changed(record)
        record.validate(softvalidation=True)
        record.set_modified(self.name)

    def get_size(self, record):
        result = record.value.get(self.name) or 0
        if isinstance(result, _FileCache):
            result = os.stat(result.path).st_size
        elif isinstance(result, (str, bytes)):
            result = len(result)
        return result

    def get_data(self, record):
        if not isinstance(record.value.get(self.name),
                (str, bytes, _FileCache)):
            if record.id < 0:
                return b''
            context = record.get_context()
            try:
                values, = RPCExecute('model', record.model_name, 'read',
                    [record.id], [self.name], context=context)
            except RPCException:
                return b''
            self._set_file_cache(record, values[self.name] or b'')
        return self.get(record)

    def get_filename(self, record, suffix=None):
        data = self.get_data(record)
        file_cache = record.value.get(self.name)
        if not isinstance(file_cache, _FileCache):
            file_cache = self._set_file_cache(record, data)
        if suffix:
            filename = file_cache.with_suffix(suffix)
        else:
            filename = file_cache.path
        return filename


class DictField(Field):

    _default = {}
    _single_value = False

    def __init__(self, attrs):
        super(DictField, self).__init__(attrs)
        self.keys = {}

    def get(self, record):
        return (super().get(record) or self._default).copy()

    def get_client(self, record):
        return super().get_client(record)

    def validation_domains(self, record, pre_validate=None):
        screen_domain, attr_domain = self.domains_get(record, pre_validate)
        return screen_domain

    def domain_get(self, record):
        screen_domain, attr_domain = self.domains_get(record)
        return concat(localize_domain(screen_domain), attr_domain)

    def date_format(self, record):
        context = self.get_context(record)
        return common.date_format(context.get('date_format'))

    def time_format(self, record):
        return '%X'

    def add_keys(self, keys, record):
        schema_model = self.attrs['schema_model']
        context = self.get_context(record)
        domain = self.domain_get(record)
        batchlen = min(10, CONFIG['client.limit'])
        for i in range(0, len(keys), batchlen):
            sub_keys = keys[i:i + batchlen]
            try:
                values = RPCExecute('model', schema_model, 'search_get_keys',
                    [('name', 'in', sub_keys), domain], CONFIG['client.limit'],
                    context=context)
            except RPCException:
                values = []
            if not values:
                continue
            self.keys.update({k['name']: k for k in values})

    def add_new_keys(self, key_ids, record):
        schema_model = self.attrs['schema_model']
        context = self.get_context(record)
        try:
            new_fields = RPCExecute('model', schema_model,
                'get_keys', key_ids, context=context)
        except RPCException:
            new_fields = []

        new_keys = []
        for new_field in new_fields:
            name = new_field['name']
            new_keys.append(name)
            self.keys[name] = new_field

        return new_keys

    def validate(self, record, softvalidation=False, pre_validate=None):
        valid = super(DictField, self).validate(
            record, softvalidation, pre_validate)

        if self.attrs.get('readonly'):
            return valid

        decoder = PYSONDecoder()
        field_value = self.get_eval(record)
        domain = []
        for key in field_value:
            if key not in self.keys:
                continue
            key_domain = self.keys[key].get('domain')
            if key_domain:
                domain.append(decoder.decode(key_domain))

        valid_value = eval_domain(domain, field_value)
        if not valid_value:
            self.get_state_attrs(record)['invalid'] = 'domain'

        return valid and valid_value


TYPES = {
    'char': CharField,
    'integer': IntegerField,
    'float': FloatField,
    'numeric': NumericField,
    'many2one': M2OField,
    'many2many': M2MField,
    'one2many': O2MField,
    'reference': ReferenceField,
    'selection': SelectionField,
    'multiselection': MultiSelectionField,
    'boolean': BooleanField,
    'datetime': DateTimeField,
    'date': DateField,
    'time': TimeField,
    'timestamp': DateTimeField,
    'timedelta': TimeDeltaField,
    'one2one': O2OField,
    'binary': BinaryField,
    'dict': DictField,
}



# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
import json
import logging
import os
from threading import Lock

from tryton.config import CONFIG, get_config_dir

logger = logging.getLogger(__name__)
COOKIES_PATH = os.path.join(get_config_dir(), 'device_cookies')


_lock = Lock()


def renew():
    from tryton.common import common

    def set_cookie(new_cookie):
        try:
            new_cookie = new_cookie()
        except Exception:
            logger.error("Cannot renew device cookie", exc_info=True)
        else:
            _set(new_cookie)

    current_cookie = get()
    common.RPCExecute(
        'model', 'res.user.device', 'renew', current_cookie,
        process_exception=False, callback=set_cookie)


def get():
    cookies = _load()
    return cookies.get(_key())


def _key():
    from tryton import common

    host = CONFIG['login.host']
    hostname = common.get_hostname(host)
    port = common.get_port(host)
    database = CONFIG['login.db']
    username = CONFIG['login.login']

    return '%(username)s@%(hostname)s:%(port)s/%(database)s' % {
        'username': username,
        'hostname': hostname,
        'port': port,
        'database': database,
        }


def _set(cookie):
    cookies = _load()
    cookies[_key()] = cookie
    try:
        with _lock:
            with open(COOKIES_PATH, 'w') as cookies_file:
                json.dump(cookies, cookies_file)
    except Exception:
        logger.error('Unable to save cookies file')


def _load():
    if not os.path.isfile(COOKIES_PATH):
        return {}
    try:
        with open(COOKIES_PATH) as cookies:
            cookies = json.load(cookies)
    except Exception:
        logger.error("Unable to load device cookies file", exc_info=True)
        cookies = {}
    return cookies



# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.

import gettext

from gi.repository import Gtk, Pango

import tryton.common as common
from tryton.common.widget_style import widget_class
from tryton.config import CONFIG
from tryton.gui import Main

from .infobar import InfoBar

_ = gettext.gettext


class ToolbarItem(object):
    def __init__(self, id, label,
            tooltip=None, icon_name=None, accel_path=None, toggle=False):
        self.id = id
        self.label = label
        self.tooltip = tooltip
        self.icon_name = icon_name
        self.accel_path = accel_path
        self.toggle = toggle

    @property
    def menu(self):
        return True

    @property
    def toolbar(self):
        return bool(self.tooltip)


class TabContent(InfoBar):

    def __init__(self, **attributes):
        super(TabContent, self).__init__()
        self.attributes = attributes.copy()

    @property
    def menu_def(self):
        return [
            ToolbarItem(
                id='switch',
                label=_("_Switch View"),
                tooltip=_("Switch View"),
                icon_name='tryton-switch',
                accel_path='<tryton>/Form/Switch View'),
            ToolbarItem(
                id='previous',
                label=_("_Previous"),
                tooltip=_("Previous Record"),
                icon_name='tryton-back',
                accel_path='<tryton>/Form/Previous'),
            ToolbarItem(
                id='next',
                label=_("_Next"),
                tooltip=_("Next Record"),
                icon_name='tryton-forward',
                accel_path='<tryton>/Form/Next'),
            ToolbarItem(
                id='search',
                label=_("_Search"),
                icon_name='tryton-search',
                accel_path='<tryton>/Form/Search'),
            None,
            ToolbarItem(
                id='new',
                label=_("_New"),
                tooltip=_("Create a new record"),
                icon_name='tryton-create',
                accel_path='<tryton>/Form/New'),
            ToolbarItem(
                id='save',
                label=_("_Save"),
                tooltip=_("Save this record"),
                icon_name='tryton-save',
                accel_path='<tryton>/Form/Save'),
            ToolbarItem(
                id='reload',
                label=_("_Reload/Undo"),
                tooltip=_("Reload/Undo"),
                icon_name='tryton-refresh',
                accel_path='<tryton>/Form/Reload'),
            ToolbarItem(
                id='copy',
                label=_("_Duplicate"),
                icon_name='tryton-copy',
                accel_path='<tryton>/Form/Duplicate'),
            ToolbarItem(
                id='remove',
                label=_("_Delete..."),
                icon_name='tryton-delete',
                accel_path='<tryton>/Form/Delete'),
            None,
            ToolbarItem(
                id='logs',
                label=_("View _Logs..."),
                icon_name='tryton-log'),
            ToolbarItem(
                id='revision' if self.model in common.MODELHISTORY else None,
                label=_("Show revisions..."),
                icon_name='tryton-history'),
            None,
            ToolbarItem(
                id='attach',
                label=_("A_ttachments..."),
                tooltip=_("Add an attachment to the record"),
                icon_name='tryton-attach',
                accel_path='<tryton>/Form/Attachments',
                toggle=True),
            ToolbarItem(
                id='note',
                label=_("_Notes..."),
                tooltip=_("Add a note to the record"),
                icon_name='tryton-note',
                accel_path='<tryton>/Form/Notes'),
            ToolbarItem(
                id='action',
                label=_("_Actions..."),
                icon_name='tryton-launch',
                accel_path='<tryton>/Form/Actions'),
            ToolbarItem(
                id='relate',
                label=_("_Relate..."),
                icon_name='tryton-link',
                accel_path='<tryton>/Form/Relate'),
            None,
            ToolbarItem(
                id='print_open',
                label=_("_Report..."),
                icon_name='tryton-open',
                accel_path='<tryton>/Form/Report'),
            ToolbarItem(
                id='print',
                label=_("_Print..."),
                icon_name='tryton-print',
                accel_path='<tryton>/Form/Print'),
            ToolbarItem(
                id='email',
                label=_("_E-Mail..."),
                tooltip=_("Send an e-mail using the record"),
                icon_name='tryton-email',
                accel_path='<tryton>/Form/Email'),
            None,
            ToolbarItem(
                id='export',
                label=_("_Export Data..."),
                icon_name='tryton-export',
                accel_path='<tryton>/Form/Export Data'),
            ToolbarItem(
                id='import',
                label=_("_Import Data..."),
                icon_name='tryton-import',
                accel_path='<tryton>/Form/Import Data'),
            ToolbarItem(
                id='copy_url',
                label=_("Copy _URL..."),
                icon_name='tryton-public',
                accel_path='<tryton>/Form/Copy URL'),
            None,
            ToolbarItem(
                id='win_close',
                label=_("_Close Tab"),
                icon_name='tryton-close',
                accel_path='<tryton>/Form/Close'),
            ]

    def create_tabcontent(self):
        self.buttons = {}
        self.menu_buttons = {}
        self.tooltips = common.Tooltips()
        self.accel_group = Main().accel_group

        self.widget = Gtk.VBox(spacing=3)
        self.widget.show()

        title_box = self.make_title_bar()
        self.widget.pack_start(title_box, expand=False, fill=True, padding=3)

        self.toolbar = self.create_toolbar(self.get_toolbars())
        self.toolbar.show_all()
        self.widget.pack_start(
            self.toolbar, expand=False, fill=True, padding=0)

        self.main = Gtk.HPaned()
        self.main.show()
        self.widget.pack_start(
            self.main, expand=True, fill=True, padding=0)

        viewport = Gtk.Viewport()
        viewport.set_shadow_type(Gtk.ShadowType.NONE)
        viewport.add(self.widget_get())
        viewport.show()
        self.scrolledwindow = Gtk.ScrolledWindow()
        self.scrolledwindow.set_shadow_type(Gtk.ShadowType.NONE)
        self.scrolledwindow.set_policy(
            Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        self.scrolledwindow.add(viewport)
        self.scrolledwindow.show()
        self.main.pack1(self.scrolledwindow, resize=True, shrink=False)

        self.widget.pack_start(
            self.create_info_bar(), expand=False, fill=True, padding=0)

    def make_title_bar(self):
        tooltips = common.Tooltips()
        self.title = title = Gtk.Label(
            label=common.ellipsize(self.name, 80),
            halign=Gtk.Align.START, margin=5,
            ellipsize=Pango.EllipsizeMode.END)
        tooltips.set_tip(title, self.name)
        title.set_size_request(0, -1)  # Allow overflow
        title.show()

        menu = Gtk.MenuButton.new()
        menu.set_relief(Gtk.ReliefStyle.NONE)
        menu.set_popup(self.set_menu_form())
        menu.show()

        self.status_label = Gtk.Label(
            margin=5, halign=Gtk.Align.END)
        widget_class(self.status_label, 'status', True)
        self.status_label.show()

        hbox = Gtk.HBox()
        hbox.pack_start(title, expand=True, fill=True, padding=0)
        hbox.pack_start(self.status_label, expand=False, fill=True, padding=0)
        hbox.show()

        frame = Gtk.Frame()
        frame.set_shadow_type(Gtk.ShadowType.ETCHED_IN)
        widget_class(frame, 'window-title', True)
        frame.add(hbox)
        frame.show()

        frame_menu = Gtk.Frame()
        frame_menu.set_shadow_type(Gtk.ShadowType.ETCHED_IN)
        frame_menu.add(menu)
        frame_menu.show()

        title_box = Gtk.HBox()
        title_box.pack_start(frame_menu, expand=False, fill=True, padding=0)
        title_box.pack_start(frame, expand=True, fill=True, padding=0)
        title_box.show()
        return title_box

    def create_base_toolbar(self, toolbar):
        previous = None
        for item in self.menu_def:
            if item and item.toolbar:
                callback = getattr(self, 'sig_%s' % item.id, None)
                if not callback:
                    continue
                if item.toggle:
                    toolitem = Gtk.ToggleToolButton()
                    toolitem.connect('toggled', callback)
                else:
                    toolitem = Gtk.ToolButton()
                    toolitem.connect('clicked', callback)
                toolitem.set_icon_widget(
                    common.IconFactory.get_image(
                        item.icon_name, Gtk.IconSize.LARGE_TOOLBAR))
                toolitem.set_label(item.label)
                toolitem.set_use_underline(True)
                self.tooltips.set_tip(toolitem, item.tooltip)
                self.buttons[item.id] = toolitem
            elif not item and previous:
                toolitem = Gtk.SeparatorToolItem()
            else:
                continue
            previous = item
            toolbar.insert(toolitem, -1)

    def set_menu_form(self):
        menu_form = Gtk.Menu()
        menu_form.set_accel_group(self.accel_group)
        menu_form.set_accel_path('<tryton>/Form')
        previous = None
        for item in self.menu_def:
            if item and item.menu:
                callback = getattr(self, 'sig_%s' % item.id, None)
                if not callback:
                    continue
                menuitem = Gtk.MenuItem(
                    label=item.label,
                    use_underline=True)
                menuitem.connect('activate', callback)
                if item.accel_path:
                    menuitem.set_accel_path(item.accel_path)
                self.menu_buttons[item.id] = menuitem
            elif not item and previous:
                menuitem = Gtk.SeparatorMenuItem()
            else:
                continue
            previous = item
            menu_form.add(menuitem)

        menu_form.show_all()
        return menu_form

    def create_toolbar(self, toolbars):
        gtktoolbar = Gtk.Toolbar()
        option = CONFIG['client.toolbar']
        if option == 'default':
            gtktoolbar.set_style(False)
        elif option == 'both':
            gtktoolbar.set_style(Gtk.ToolbarStyle.BOTH)
        elif option == 'text':
            gtktoolbar.set_style(Gtk.ToolbarStyle.TEXT)
        elif option == 'icons':
            gtktoolbar.set_style(Gtk.ToolbarStyle.ICONS)
        self.create_base_toolbar(gtktoolbar)
        return gtktoolbar

    def compare(self, model, attributes):
        return False



# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
import gettext

from gi.repository import Gdk, Gtk

from tryton.common import data2pixbuf, resize_pixbuf, url_open
from tryton.config import CONFIG

from .binary import BinaryMixin
from .widget import Widget

_ = gettext.gettext


class Image(BinaryMixin, Widget):

    def __init__(self, view, attrs):
        super(Image, self).__init__(view, attrs)

        self.height = int(attrs.get('height', 100))
        self.width = int(attrs.get('width', 300))

        self.widget = Gtk.Frame()
        self.widget.set_shadow_type(Gtk.ShadowType.ETCHED_IN)
        self.widget.get_accessible().set_name(attrs.get('string', ''))
        vbox = Gtk.VBox(spacing=3, margin=5)
        self.widget.add(vbox)
        self.event = Gtk.EventBox()
        self.event.drag_dest_set(Gtk.DestDefaults.ALL, [
            Gtk.TargetEntry.new('text/plain', 0, 0),
            Gtk.TargetEntry.new('text/uri-list', 0, 1),
            Gtk.TargetEntry.new("image/x-xpixmap", 0, 2)],
            Gdk.DragAction.MOVE)
        self.event.connect('drag_motion', self.drag_motion)
        self.event.connect('drag_data_received', self.drag_data_received)

        self.image = Gtk.Image()
        self.event.add(self.image)
        vbox.pack_start(self.event, expand=True, fill=True, padding=0)

        toolbar = self.toolbar()  # Set button attributes even if not display
        if not attrs.get('readonly'):
            toolbar.set_halign(Gtk.Align.CENTER)
            vbox.pack_start(
                toolbar, expand=False, fill=False, padding=0)

        self._readonly = False

    @property
    def filters(self):
        filters = super(Image, self).filters
        filter_image = Gtk.FileFilter()
        filter_image.set_name(_('Images'))
        for mime in ("image/png", "image/jpeg", "image/gif"):
            filter_image.add_mime_type(mime)
        for pat in ("*.png", "*.jpg", "*.gif", "*.tif", "*.xpm"):
            filter_image.add_pattern(pat)
        filters.insert(0, filter_image)
        return filters

    def _readonly_set(self, value):
        self._readonly = value
        self.but_select.set_sensitive(not value)
        self.but_clear.set_sensitive(not value)

    def clear(self, widget=None):
        super(Image, self).clear(widget=widget)
        self.update_img()

    def drag_motion(self, widget, context, x, y, timestamp):
        if self._readonly:
            return False
        Gdk.drag_status(context, Gdk.DragAction.COPY, timestamp)
        return True

    def drag_data_received(self, widget, context, x, y, selection,
            info, timestamp):
        if self._readonly:
            return
        if info == 0:
            uri = selection.get_text().split('\n')[0]
            if uri:
                self.field.set_client(self.record, url_open(uri).read())
            self.update_img()
        elif info == 1:
            uri = selection.get_data().split('\r\n')[0]
            if uri:
                self.field.set_client(self.record, url_open(uri).read())
            self.update_img()
        elif info == 2:
            data = selection.get_pixbuf()
            if data:
                self.field.set_client(self.record, data)
                self.update_img()

    def update_img(self):
        value = None
        if self.field:
            value = self.field.get_client(self.record)
        if isinstance(value, int):
            if value > CONFIG['image.max_size']:
                value = False
            else:
                value = self.field.get_data(self.record)
        pixbuf = data2pixbuf(value)
        if pixbuf:
            pixbuf = resize_pixbuf(pixbuf, self.width, self.height)
        self.image.set_from_pixbuf(pixbuf)
        return bool(value)

    def display(self):
        super(Image, self).display()
        value = self.update_img()
        self.update_buttons(bool(value))



# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
import gettext
import sys

from gi.repository import Gtk

from tryton.common import (
    IconFactory, file_selection, get_toplevel_window, message, node_attributes)
from tryton.common.underline import set_underline
from tryton.config import TRYTON_ICON
from tryton.gui import Main

from . import View, XMLViewParser
from .graph_gtk.bar import HorizontalBar, VerticalBar
from .graph_gtk.line import Line
from .graph_gtk.pie import Pie

_ = gettext.gettext


class GraphXMLViewParser(XMLViewParser):

    WIDGETS = {
        'hbar': HorizontalBar,
        'line': Line,
        'pie': Pie,
        'vbar': VerticalBar,
        }

    def __init__(self, view, exclude_field, field_attrs):
        super().__init__(view, exclude_field, field_attrs)
        self._xfield = None
        self._yfields = []

    def _node_attributes(self, node):
        node_attrs = node_attributes(node)
        if 'name' in node_attrs:
            if not node_attrs.get('string') and node_attrs['name'] != '#':
                field = self.field_attrs[node_attrs['name']]
                node_attrs['string'] = field['string']
        return node_attrs

    def _parse_graph(self, node, attributes):
        for child in node.childNodes:
            self.parse(child)

        Widget = self.WIDGETS.get(attributes.get('type', 'vbar'))
        widget = Widget(
            self.view, self._xfield, self._yfields)
        self.view.widget.add(widget)
        self.view.widgets['root'] = widget

    def _parse_x(self, node, attributes):
        for child in node.childNodes:
            if not child.nodeType == child.ELEMENT_NODE:
                continue
            self._xfield = self._node_attributes(child)

    def _parse_y(self, node, attributes):
        for child in node.childNodes:
            if not child.nodeType == child.ELEMENT_NODE:
                continue
            self._yfields.append(self._node_attributes(child))


class ViewGraph(View):
    view_type = 'graph'
    editable = False
    creatable = False
    xml_parser = GraphXMLViewParser

    def __init__(self, view_id, screen, xml):
        self.widget = event = Gtk.EventBox()
        super().__init__(view_id, screen, xml)
        event.connect('button-press-event', self.button_press)

    def __getitem__(self, name):
        return None

    def destroy(self):
        self.widget.destroy()
        self.widgets['root'].destroy()
        self.widgets.clear()

    def set_value(self):
        pass

    def reset(self):
        pass

    def display(self):
        self.widgets['root'].display(self.screen.group)
        return True

    def set_cursor(self, new=False, reset_view=True):
        pass

    def get_fields(self):
        return []

    def get_buttons(self):
        return []

    def save(self, widget):
        parent = get_toplevel_window()
        dia = Gtk.Dialog(
            title=_('Image Size'), transient_for=parent, modal=True,
            destroy_with_parent=True)
        Main().add_window(dia)
        cancel_button = dia.add_button(
            set_underline(_("Cancel")), Gtk.ResponseType.CANCEL)
        cancel_button.set_image(IconFactory.get_image(
                'tryton-cancel', Gtk.IconSize.BUTTON))
        cancel_button.set_always_show_image(True)
        ok_button = dia.add_button(
            set_underline(_("OK")), Gtk.ResponseType.OK)
        ok_button.set_image(IconFactory.get_image(
                'tryton-ok', Gtk.IconSize.BUTTON))
        ok_button.set_always_show_image(True)
        dia.set_icon(TRYTON_ICON)
        dia.set_default_response(Gtk.ResponseType.OK)

        hbox = Gtk.HBox(spacing=3)
        dia.vbox.pack_start(hbox, expand=False, fill=True, padding=0)

        hbox.pack_start(
            Gtk.Label(label=_('Width:')), expand=False, fill=True, padding=0)
        spinwidth = Gtk.SpinButton()
        spinwidth.configure(Gtk.Adjustment(
                value=400.0, lower=0.0, upper=sys.maxsize,
                step_increment=1.0, page_increment=10.0),
            climb_rate=1, digits=0)
        spinwidth.set_numeric(True)
        spinwidth.set_activates_default(True)
        hbox.pack_start(spinwidth, expand=True, fill=True, padding=0)

        hbox.pack_start(
            Gtk.Label(label=_('Height:')), expand=False, fill=True, padding=0)
        spinheight = Gtk.SpinButton()
        spinheight.configure(Gtk.Adjustment(
                value=200.0, lower=0.0, upper=sys.maxsize,
                step_increment=1.0, page_increment=10.0),
            climb_rate=1, digits=0)
        spinheight.set_numeric(True)
        spinheight.set_activates_default(True)
        hbox.pack_start(spinheight, expand=True, fill=True, padding=0)
        dia.show_all()

        filter = Gtk.FileFilter()
        filter.set_name(_('PNG image (*.png)'))
        filter.add_mime_type('image/png')
        filter.add_pattern('*.png')

        while True:
            response = dia.run()
            width = spinwidth.get_value_as_int()
            height = spinheight.get_value_as_int()
            if response == Gtk.ResponseType.OK:
                filename = file_selection(
                    _('Save As'),
                    action=Gtk.FileChooserAction.SAVE,
                    preview=False,
                    filters=[filter])
                if width and height and filename:
                    filename = filename.with_suffix('.png')
                    try:
                        self.widgets['root'].export_png(
                            filename, width, height)
                        break
                    except MemoryError:
                        message(
                            _('Image size too large.'), dia,
                            Gtk.MessageType.ERROR)
            else:
                break
        parent.present()
        dia.destroy()

    def button_press(self, widget, event):
        if event.button == 3:
            menu = Gtk.Menu()
            item = Gtk.MenuItem(label=_('Save As...'))
            item.connect('activate', self.save)
            item.show()
            menu.append(item)
            if hasattr(menu, 'popup_at_pointer'):
                menu.popup_at_pointer(event)
            else:
                menu.popup(None, None, None, event.button, event.time)
            return True
        elif event.button == 1:
            self.widgets['root'].action()



# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
"Attachment"
import gettext
import os
import webbrowser
from functools import partial
from urllib.parse import unquote, urlparse

from tryton.common import (
    RPCException, RPCExecute, file_open, file_write, url_open)
from tryton.gui.window.view_form.screen import Screen
from tryton.gui.window.win_form import WinForm

_ = gettext.gettext


class Attachment(WinForm):
    "Attachment window"

    def __init__(self, record, callback=None):
        self.resource = '%s,%s' % (record.model_name, record.id)
        self.attachment_callback = callback
        title = _('Attachments (%s)') % (record.rec_name())
        screen = Screen('ir.attachment', domain=[
            ('resource', '=', self.resource),
            ], mode=['tree', 'form'])
        super(Attachment, self).__init__(screen, self.callback,
            view_type='tree', title=title)
        screen.search_filter()

    def destroy(self):
        self.prev_view.save_width()
        super(Attachment, self).destroy()

    def callback(self, result):
        if result:
            self.screen.save_current()
        if self.attachment_callback:
            self.attachment_callback()

    def add_uri(self, uri):
        data_field = self.screen.group.fields['data']
        link_field = self.screen.group.fields['link']
        type_field = self.screen.group.fields['type']
        name_field = self.screen.group.fields[data_field.attrs['filename']]
        description_field = self.screen.group.fields['description']
        new_record = self.screen.new()
        parse = urlparse(uri)
        if not parse.scheme:
            description_field.set_client(new_record, uri)
        else:
            file_name = os.path.basename(unquote(parse.path))
            name_field.set_client(new_record, file_name)
            if parse.scheme == 'file':
                data_field.set_client(new_record, url_open(uri).read())
                type_field.set_client(new_record, 'data')
            else:
                link_field.set_client(new_record, uri)
                type_field.set_client(new_record, 'link')
        self.screen.display()

    def add_file(self, filename):
        self.add_uri(filename.as_uri())

    @staticmethod
    def get_attachments(record):
        attachments = []
        context = {}
        if record and record.id >= 0:
            context = record.get_context()
            try:
                attachments = RPCExecute('model', 'ir.attachment',
                    'search_read', [
                        ('resource', '=', '%s,%s' % (
                                record.model_name, record.id)),
                        ], 0, 20, None, ['rec_name', 'name', 'type', 'link'],
                    context=context)
            except RPCException:
                pass
        for attachment in attachments:
            name = attachment['rec_name']
            callback = getattr(
                Attachment, 'open_' + attachment['type'], Attachment.open_data)
            yield name, partial(
                callback, attachment=attachment, context=context)

    @staticmethod
    def open_link(attachment, context):
        if attachment['link']:
            webbrowser.open(attachment['link'], new=2)

    @staticmethod
    def open_data(attachment, context):
        try:
            value, = RPCExecute('model', 'ir.attachment', 'read',
                [attachment['id']], ['data'], context=context)
        except RPCException:
            return
        filepath = file_write(attachment['name'], value['data'])
        file_open(filepath)



# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
import tryton.common as common


class NoModal(object):

    def __init__(self):
        self.parent = common.get_toplevel_window()
        self.sensible_widget = common.get_sensible_widget(self.parent)
        self.page = None
        self.parent_focus = []
        focus = self.parent.get_focus()
        while focus:
            self.parent_focus.append(focus)
            focus = focus.get_parent()

    def register(self):
        from tryton.gui.main import Main
        main = Main()
        self.page = main.get_page()
        if not self.page:
            self.page = main
        self.page.dialogs.append(self)
        self.sensible_widget.props.sensitive = False

    def destroy(self):
        if not self.page:
            return
        self.page.dialogs.remove(self)
        self.parent.present()
        self.sensible_widget.props.sensitive = True
        for focus in self.parent_focus:
            if focus and focus.is_ancestor(self.parent):
                try:
                    focus.grab_focus()
                except TypeError:
                    # GooCanvas needs a GooCanvasItem
                    continue
                break

    def show(self):
        raise NotImplementedError

    def hide(self):
        raise NotImplementedError

    def default_size(self):
        from tryton.gui.main import Main
        main = Main()
        allocation = main.window.get_allocation()
        width, height = allocation.width, allocation.height
        return max(width - 150, 0), max(height - 150, 0)



# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
import gettext
import locale
import os
import sys

from gi.repository import Gdk, GObject, Gtk

from tryton.common import IconFactory
from tryton.common.underline import set_underline
from tryton.config import TRYTON_ICON
from tryton.gui import Main
from tryton.gui.window.nomodal import NoModal

_ = gettext.gettext

encodings = ["ascii", "big5", "big5hkscs", "cp037", "cp424", "cp437", "cp500",
    "cp720", "cp737", "cp775", "cp850", "cp852", "cp855", "cp856", "cp857",
    "cp858", "cp860", "cp861", "cp862", "cp863", "cp864", "cp865", "cp866",
    "cp869", "cp874", "cp875", "cp932", "cp949", "cp950", "cp1006", "cp1026",
    "cp1140", "cp1250", "cp1251", "cp1252", "cp1253", "cp1254", "cp1255",
    "cp1256", "cp1257", "cp1258", "euc_jp", "euc_jis_2004", "euc_jisx0213",
    "euc_kr", "gb2312", "gbk", "gb18030", "hz", "iso2022_jp", "iso2022_jp_1",
    "iso2022_jp_2", "iso2022_jp_2004", "iso2022_jp_3", "iso2022_jp_ext",
    "iso2022_kr", "latin_1", "iso8859_2", "iso8859_3", "iso8859_4",
    "iso8859_5", "iso8859_6", "iso8859_7", "iso8859_8", "iso8859_9",
    "iso8859_10", "iso8859_13", "iso8859_14", "iso8859_15", "iso8859_16",
    "johab", "koi8_r", "koi8_u", "mac_cyrillic", "mac_greek", "mac_iceland",
    "mac_latin2", "mac_roman", "mac_turkish", "ptcp154", "shift_jis",
    "shift_jis_2004", "shift_jisx0213", "utf_32", "utf_32_be", "utf_32_le",
    "utf_16", "utf_16_be", "utf_16_le", "utf_7", "utf_8", "utf_8_sig"]


class WinCSV(NoModal):
    def __init__(self, *args, **kwargs):
        super(WinCSV, self).__init__(*args, **kwargs)

        self.dialog = Gtk.Dialog(
            transient_for=self.parent, destroy_with_parent=True)
        Main().add_window(self.dialog)
        self.dialog.set_position(Gtk.WindowPosition.CENTER_ON_PARENT)
        self.dialog.set_icon(TRYTON_ICON)
        self.dialog.set_default_size(*self.default_size())
        self.dialog.connect('response', self.response)

        dialog_vbox = Gtk.VBox()

        hbox_mapping = Gtk.HBox(homogeneous=True)
        dialog_vbox.pack_start(hbox_mapping, expand=True, fill=True, padding=0)

        frame_fields = Gtk.Frame()
        frame_fields.set_shadow_type(Gtk.ShadowType.NONE)
        viewport_fields = Gtk.Viewport()
        scrolledwindow_fields = Gtk.ScrolledWindow()
        scrolledwindow_fields.set_policy(
            Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        viewport_fields.add(scrolledwindow_fields)
        frame_fields.add(viewport_fields)
        label_all_fields = Gtk.Label(
            label=_('<b>All fields</b>'), use_markup=True)
        frame_fields.set_label_widget(label_all_fields)
        hbox_mapping.pack_start(
            frame_fields, expand=True, fill=True, padding=0)

        vbox_buttons = Gtk.VBox(homogeneous=False, spacing=10)
        vbox_buttons.set_border_width(5)
        hbox_mapping.pack_start(
            vbox_buttons, expand=False, fill=True, padding=0)

        button_add = Gtk.Button(
            label=_('_Add'), stock=None, use_underline=True)
        button_add.set_image(IconFactory.get_image(
                'tryton-add', Gtk.IconSize.BUTTON))
        button_add.set_always_show_image(True)
        button_add.connect_after('clicked', self.sig_sel)
        vbox_buttons.pack_start(
            button_add, expand=False, fill=False, padding=0)

        button_remove = Gtk.Button(
            label=_('_Remove'), stock=None, use_underline=True)
        button_remove.set_image(IconFactory.get_image(
                'tryton-remove', Gtk.IconSize.BUTTON))
        button_remove.set_always_show_image(True)
        button_remove.connect_after('clicked', self.sig_unsel)
        vbox_buttons.pack_start(
            button_remove, expand=False, fill=False, padding=0)

        button_remove_all = Gtk.Button(
            label=_('_Clear'), stock=None, use_underline=True)
        button_remove_all.set_image(IconFactory.get_image(
                'tryton-clear', Gtk.IconSize.BUTTON))
        button_remove_all.set_always_show_image(True)
        button_remove_all.connect_after('clicked', self.sig_unsel_all)
        vbox_buttons.pack_start(
            button_remove_all, expand=False, fill=False, padding=0)

        hseparator_buttons = Gtk.HSeparator()
        vbox_buttons.pack_start(
            hseparator_buttons, expand=False, fill=False, padding=3)

        self.add_buttons(vbox_buttons)

        frame_fields_selected = Gtk.Frame()
        frame_fields_selected.set_shadow_type(Gtk.ShadowType.NONE)
        viewport_fields_selected = Gtk.Viewport()
        scrolledwindow_fields_selected = Gtk.ScrolledWindow()
        scrolledwindow_fields_selected.set_policy(
            Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        viewport_fields_selected.add(scrolledwindow_fields_selected)
        frame_fields_selected.add(viewport_fields_selected)
        label_fields_selected = Gtk.Label(
            label=_('<b>Fields selected</b>'), use_markup=True)
        frame_fields_selected.set_label_widget(label_fields_selected)
        hbox_mapping.pack_start(
            frame_fields_selected, expand=True, fill=True, padding=0)

        vbox_csv_param = Gtk.VBox()
        vbox_csv_param.props.margin = 7
        dialog_vbox.pack_start(
            vbox_csv_param, expand=False, fill=True, padding=0)

        self.add_chooser(vbox_csv_param)

        expander_csv = Gtk.Expander()
        vbox_csv_param.pack_start(
            expander_csv, expand=False, fill=True, padding=0)
        label_csv_param = Gtk.Label(label=_('CSV Parameters'))
        expander_csv.set_label_widget(label_csv_param)

        box = Gtk.HBox(spacing=3)
        expander_csv.add(box)

        label_csv_delimiter = Gtk.Label(
            label=_('Delimiter:'), halign=Gtk.Align.END)
        box.pack_start(label_csv_delimiter, expand=False, fill=True, padding=0)
        self.csv_delimiter = Gtk.Entry()
        self.csv_delimiter.set_max_length(1)
        if os.name == 'nt' and ',' == locale.localeconv()['decimal_point']:
            delimiter = ';'
        else:
            delimiter = ','
        self.csv_delimiter.set_text(delimiter)
        self.csv_delimiter.set_width_chars(1)
        label_csv_delimiter.set_mnemonic_widget(self.csv_delimiter)
        box.pack_start(
            self.csv_delimiter, expand=False, fill=True, padding=0)

        label_csv_quotechar = Gtk.Label(
            label=_("Quote char:"), halign=Gtk.Align.END)
        box.pack_start(label_csv_quotechar, expand=False, fill=True, padding=0)
        self.csv_quotechar = Gtk.Entry()
        self.csv_quotechar.set_text("\"")
        self.csv_quotechar.set_width_chars(1)
        label_csv_quotechar.set_mnemonic_widget(self.csv_quotechar)
        box.pack_start(self.csv_quotechar, expand=False, fill=True, padding=0)

        label_csv_enc = Gtk.Label(
            label=_("Encoding:"), halign=Gtk.Align.END)
        box.pack_start(label_csv_enc, expand=False, fill=True, padding=0)
        self.csv_enc = Gtk.ComboBoxText()
        for i, encoding in enumerate(encodings):
            self.csv_enc.append_text(encoding)
            if encoding == 'utf_8_sig':
                self.csv_enc.set_active(i)
        label_csv_enc.set_mnemonic_widget(self.csv_enc)
        box.pack_start(self.csv_enc, expand=False, fill=True, padding=0)

        self.csv_locale = Gtk.CheckButton(label=_("Use locale format"))
        self.csv_locale.set_active(True)
        box.pack_start(self.csv_locale, expand=False, fill=True, padding=0)

        self.add_csv_header_param(box)

        button_cancel = self.dialog.add_button(
            set_underline(_("Cancel")), Gtk.ResponseType.CANCEL)
        button_cancel.set_image(IconFactory.get_image(
                'tryton-cancel', Gtk.IconSize.BUTTON))

        button_ok = self.dialog.add_button(
            set_underline(_("OK")), Gtk.ResponseType.OK)
        button_ok.set_image(IconFactory.get_image(
                'tryton-ok', Gtk.IconSize.BUTTON))

        self.dialog.vbox.pack_start(
            dialog_vbox, expand=True, fill=True, padding=0)

        self.view1 = Gtk.TreeView()
        self.view1.get_selection().set_mode(Gtk.SelectionMode.MULTIPLE)
        self.view1.connect('row-expanded', self.on_row_expanded)
        scrolledwindow_fields.add(self.view1)
        self.view2 = Gtk.TreeView()
        self.view2.get_selection().set_mode(Gtk.SelectionMode.MULTIPLE)
        scrolledwindow_fields_selected.add(self.view2)
        self.view1.set_headers_visible(False)
        self.view2.set_headers_visible(False)

        cell = Gtk.CellRendererText()
        column = Gtk.TreeViewColumn(_('Field name'), cell, text=0)
        self.view1.append_column(column)

        cell = Gtk.CellRendererText()
        column = Gtk.TreeViewColumn(_('Field name'), cell, text=0)
        self.view2.append_column(column)

        self.model1 = Gtk.TreeStore(GObject.TYPE_STRING, GObject.TYPE_STRING)
        self.model2 = Gtk.ListStore(GObject.TYPE_STRING, GObject.TYPE_STRING)

        self.model_populate(self._get_fields(self.model))

        self.view1.set_model(self.model1)
        self.view1.connect('row-activated', self.sig_sel)
        self.view2.set_model(self.model2)
        self.view2.connect('row-activated', self.sig_unsel)

        self.dialog.show_all()
        self.show()

        self.register()

        if sys.platform != 'darwin':
            self.view2.drag_source_set(
                Gdk.ModifierType.BUTTON1_MASK | Gdk.ModifierType.BUTTON3_MASK,
                [Gtk.TargetEntry.new(
                        'EXPORT_TREE', Gtk.TargetFlags.SAME_WIDGET, 0)],
                Gdk.DragAction.MOVE)
            self.view2.drag_dest_set(
                Gtk.DestDefaults.ALL,
                [Gtk.TargetEntry.new(
                        'EXPORT_TREE', Gtk.TargetFlags.SAME_WIDGET, 0)],
                Gdk.DragAction.MOVE)
            self.view2.connect('drag-begin', self.drag_begin)
            self.view2.connect('drag-motion', self.drag_motion)
            self.view2.connect('drag-drop', self.drag_drop)
            self.view2.connect("drag-data-get", self.drag_data_get)
            self.view2.connect('drag-data-received', self.drag_data_received)
            self.view2.connect('drag-data-delete', self.drag_data_delete)

            drag_column = Gtk.TreeViewColumn()
            drag_column.set_sizing(Gtk.TreeViewColumnSizing.FIXED)
            cell_pixbuf = Gtk.CellRendererPixbuf()
            cell_pixbuf.props.pixbuf = IconFactory.get_pixbuf('tryton-drag')
            drag_column.pack_start(cell_pixbuf, expand=False)
            self.view2.insert_column(drag_column, 0)

    def drag_begin(self, treeview, context):
        return True

    def drag_motion(self, treeview, context, x, y, time):
        try:
            treeview.set_drag_dest_row(*treeview.get_dest_row_at_pos(x, y))
        except TypeError:
            treeview.set_drag_dest_row(
                Gtk.TreePath(len(treeview.get_model()) - 1),
                Gtk.TreeViewDropPosition.AFTER)
        Gdk.drag_status(context, Gdk.DragAction.MOVE, time)
        return True

    def drag_drop(self, treeview, context, x, y, time):
        treeview.stop_emission_by_name('drag-drop')
        return True

    def drag_data_get(self, treeview, context, selection, target_id,
            etime):
        treeview.stop_emission_by_name('drag-data-get')

        def _func_sel_get(store, path, iter_, data):
            data.append(path[0])
        data = []
        treeselection = treeview.get_selection()
        treeselection.selected_foreach(_func_sel_get, data)
        if not data:
            return
        data = ','.join(str(x) for x in data)
        selection.set(selection.get_target(), 8, data.encode('utf-8'))
        return True

    def drag_data_received(self, treeview, context, x, y, selection,
            info, etime):
        treeview.stop_emission_by_name('drag-data-received')
        selection_data = selection.get_data()
        if not selection_data:
            return
        selection_data = selection_data.decode('utf-8')
        store = treeview.get_model()

        data_iters = [store.get_iter((int(i),))
            for i in selection_data.split(',')]
        drop_info = treeview.get_dest_row_at_pos(x, y)
        if drop_info:
            path, position = drop_info
            pos = store.get_iter(path)
        else:
            pos = store.get_iter((len(store) - 1,))
            position = Gtk.TreeViewDropPosition.AFTER
        if position == Gtk.TreeViewDropPosition.AFTER:
            data_iters = reversed(data_iters)
        for item in data_iters:
            if position == Gtk.TreeViewDropPosition.BEFORE:
                store.move_before(item, pos)
            else:
                store.move_after(item, pos)
        Gdk.drop_finish(context, False, etime)
        return True

    def drag_data_delete(self, treeview, context):
        treeview.stop_emission_by_name('drag-data-delete')

    def get_delimiter(self):
        return self.csv_delimiter.get_text() or ','

    def get_quotechar(self):
        return self.csv_quotechar.get_text() or '"'

    def get_encoding(self):
        return self.csv_enc.get_active_text() or 'utf_8'

    def destroy(self):
        super(WinCSV, self).destroy()
        self.dialog.destroy()

    def show(self):
        self.dialog.show()

    def hide(self):
        self.dialog.hide()



# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
from .view_board import ViewBoard

__all__ = [ViewBoard]
